# ruff: noqa: N802
import ast
import builtins
import functools
import inspect
from collections.abc import Callable, Sequence
from inspect import ismethod
from types import FunctionType, MethodType, MethodWrapperType
from typing import Any, Never, Self

from sonolus.backend.excepthook import install_excepthook
from sonolus.backend.utils import get_function, scan_writes
from sonolus.script.debug import assert_true
from sonolus.script.internal.builtin_impls import BUILTIN_IMPLS, _bool, _float, _int, _len
from sonolus.script.internal.constant import ConstantValue
from sonolus.script.internal.context import Context, EmptyBinding, Scope, ValueBinding, ctx, set_ctx
from sonolus.script.internal.descriptor import SonolusDescriptor
from sonolus.script.internal.error import CompilationError
from sonolus.script.internal.impl import validate_value
from sonolus.script.internal.tuple_impl import TupleImpl
from sonolus.script.internal.value import Value
from sonolus.script.iterator import SonolusIterator
from sonolus.script.num import Num, _is_num

_compiler_internal_ = True


def compile_and_call[**P, R](fn: Callable[P, R], /, *args: P.args, **kwargs: P.kwargs) -> R:
    if not ctx():
        return fn(*args, **kwargs)
    return validate_value(generate_fn_impl(fn)(*args, **kwargs))


def generate_fn_impl(fn: Callable):
    install_excepthook()
    match fn:
        case ConstantValue() as value if value._is_py_():
            return generate_fn_impl(value._as_py_())
        case MethodType() as method:
            return functools.partial(generate_fn_impl(method.__func__), method.__self__)
        case FunctionType() as function:
            if getattr(function, "_meta_fn_", False):
                return function
            return functools.partial(eval_fn, function)
        case _:
            if callable(fn) and isinstance(fn, Value):
                return generate_fn_impl(fn.__call__)
            elif callable(fn):
                raise TypeError(f"Unsupported callable {fn!r}")
            else:
                raise TypeError(f"'{type(fn).__name__}' object is not callable")


def eval_fn(fn: Callable, /, *args, **kwargs):
    source_file, node = get_function(fn)
    bound_args = inspect.signature(fn).bind(*args, **kwargs)
    bound_args.apply_defaults()
    if ismethod(fn):
        code = fn.__func__.__code__
        closure = fn.__func__.__closure__
    else:
        code = fn.__code__
        closure = fn.__closure__
    if closure is None:
        nonlocal_vars = {}
    else:
        nonlocal_vars = {
            var: cell.cell_contents for var, cell in zip(code.co_freevars, closure, strict=True) if cell is not None
        }
    global_vars = {
        **builtins.__dict__,
        **fn.__globals__,
        **nonlocal_vars,
    }
    return Visitor(source_file, bound_args, global_vars).run(node)


unary_ops = {
    ast.Invert: "__invert__",
    ast.UAdd: "__pos__",
    ast.USub: "__neg__",
}

bin_ops = {
    ast.Add: "__add__",
    ast.Sub: "__sub__",
    ast.Mult: "__mul__",
    ast.Div: "__truediv__",
    ast.FloorDiv: "__floordiv__",
    ast.Mod: "__mod__",
    ast.Pow: "__pow__",
    ast.LShift: "__lshift__",
    ast.RShift: "__rshift__",
    ast.BitOr: "__or__",
    ast.BitAnd: "__and__",
    ast.BitXor: "__xor__",
    ast.MatMult: "__matmul__",
}

rbin_ops = {
    ast.Add: "__radd__",
    ast.Sub: "__rsub__",
    ast.Mult: "__rmul__",
    ast.Div: "__rtruediv__",
    ast.FloorDiv: "__rfloordiv__",
    ast.Mod: "__rmod__",
    ast.Pow: "__rpow__",
    ast.LShift: "__rlshift__",
    ast.RShift: "__rrshift__",
    ast.BitOr: "__ror__",
    ast.BitAnd: "__rand__",
    ast.BitXor: "__rxor__",
    ast.MatMult: "__rmatmul__",
}

inplace_ops = {
    ast.Add: "__iadd__",
    ast.Sub: "__isub__",
    ast.Mult: "__imul__",
    ast.Div: "__itruediv__",
    ast.FloorDiv: "__ifloordiv__",
    ast.Mod: "__imod__",
    ast.Pow: "__ipow__",
    ast.LShift: "__ilshift__",
    ast.RShift: "__irshift__",
    ast.BitOr: "__ior__",
    ast.BitXor: "__ixor__",
    ast.BitAnd: "__iand__",
    ast.MatMult: "__imatmul__",
}

comp_ops = {
    ast.Eq: "__eq__",
    ast.NotEq: "__ne__",
    ast.Lt: "__lt__",
    ast.LtE: "__le__",
    ast.Gt: "__gt__",
    ast.GtE: "__ge__",
}

rcomp_ops = {
    ast.Eq: "__eq__",
    ast.NotEq: "__ne__",
    ast.Lt: "__gt__",
    ast.LtE: "__ge__",
    ast.Gt: "__lt__",
    ast.GtE: "__le__",
    ast.In: "__contains__",  # Only supported on the right side
    ast.NotIn: "__contains__",
}

op_to_symbol = {
    ast.Add: "+",
    ast.Sub: "-",
    ast.Mult: "*",
    ast.Div: "/",
    ast.FloorDiv: "//",
    ast.Mod: "%",
    ast.Pow: "**",
    ast.Eq: "==",
    ast.NotEq: "!=",
    ast.Lt: "<",
    ast.LtE: "<=",
    ast.Gt: ">",
    ast.GtE: ">=",
    ast.And: "and",
    ast.Or: "or",
    ast.BitAnd: "&",
    ast.BitOr: "|",
    ast.BitXor: "^",
    ast.LShift: "<<",
    ast.RShift: ">>",
    ast.USub: "-",
    ast.UAdd: "+",
    ast.Invert: "~",
    ast.Not: "not",
    ast.In: "in",
    ast.NotIn: "not in",
}


class Visitor(ast.NodeVisitor):
    source_file: str
    globals: dict[str, Any]
    bound_args: inspect.BoundArguments
    used_names: dict[str, int]
    return_ctxs: list[Context]  # Contexts at return statements, which will branch to the exit
    loop_head_ctxs: list[
        Context | list[Context]
    ]  # Contexts at loop heads, from outer to inner. Contains a list for unrolled (tuple) loops
    break_ctxs: list[list[Context]]  # Contexts at break statements, from outer to inner
    active_ctx: Context | None  # The active context for use in nested functions=
    parent: Self | None  # The parent visitor for use in nested functions

    def __init__(
        self,
        source_file: str,
        bound_args: inspect.BoundArguments,
        global_vars: dict[str, Any],
        parent: Self | None = None,
    ):
        self.source_file = source_file
        self.globals = global_vars
        self.bound_args = bound_args
        self.used_names = {}
        self.return_ctxs = []
        self.loop_head_ctxs = []
        self.break_ctxs = []
        self.active_ctx = None
        self.parent = parent

    def run(self, node):
        before_ctx = ctx()
        set_ctx(before_ctx.branch_with_scope(None, Scope()))
        for name, value in self.bound_args.arguments.items():
            ctx().scope.set_value(name, validate_value(value))
        match node:
            case ast.FunctionDef(body=body):
                ctx().scope.set_value("$return", validate_value(None))
                for stmt in body:
                    if not ctx().live:
                        break
                    self.visit(stmt)
            case ast.Lambda(body=body):
                result = self.visit(body)
                ctx().scope.set_value("$return", result)
            case _:
                raise NotImplementedError("Unsupported syntax")
        after_ctx = Context.meet([*self.return_ctxs, ctx()])
        self.active_ctx = after_ctx
        result_binding = after_ctx.scope.get_binding("$return")
        if not isinstance(result_binding, ValueBinding):
            raise ValueError("Function has conflicting return values")
        set_ctx(after_ctx.branch_with_scope(None, before_ctx.scope.copy()))
        return result_binding.value

    def visit(self, node):
        """Visit a node."""
        # We want this here so this is filtered out of tracebacks
        method = "visit_" + node.__class__.__name__
        visitor = getattr(self, method, self.generic_visit)
        with self.reporting_errors_at_node(node):
            return visitor(node)

    def visit_FunctionDef(self, node):
        name = node.name
        signature = self.arguments_to_signature(node.args)

        def fn(*args, **kwargs):
            bound = signature.bind(*args, **kwargs)
            bound.apply_defaults()
            return Visitor(
                self.source_file,
                bound,
                self.globals,
                self,
            ).run(node)

        fn._meta_fn_ = True
        fn.__name__ = name
        fn.__qualname__ = name

        for decorator in reversed(node.decorator_list):
            fn = self.handle_call(decorator, self.visit(decorator), fn)

        ctx().scope.set_value(name, validate_value(fn))

    def visit_AsyncFunctionDef(self, node):
        raise NotImplementedError("Async functions are not supported")

    def visit_ClassDef(self, node):
        raise NotImplementedError("Classes within functions are not supported")

    def visit_Return(self, node):
        value = self.visit(node.value) if node.value else validate_value(None)
        ctx().scope.set_value("$return", value)
        self.return_ctxs.append(ctx())
        set_ctx(ctx().into_dead())

    def visit_Delete(self, node):
        for target in node.targets:
            match target:
                case ast.Name():
                    raise NotImplementedError("Deleting variables is not supported")
                case ast.Subscript(value=value, slice=slice):
                    self.handle_delitem(target, self.visit(value), self.visit(slice))
                case ast.Attribute():
                    raise NotImplementedError("Deleting attributes is not supported")
                case _:
                    raise NotImplementedError("Unsupported delete target")

    def visit_Assign(self, node):
        value = self.visit(node.value)
        for target in node.targets:
            self.handle_assign(target, value)

    def visit_TypeAlias(self, node):
        raise NotImplementedError("Type aliases are not supported")

    def visit_AugAssign(self, node):
        lhs_value = self.visit(node.target)
        rhs_value = self.visit(node.value)
        inplace_fn_name = inplace_ops[type(node.op)]
        regular_fn_name = bin_ops[type(node.op)]
        right_fn_name = rbin_ops[type(node.op)]
        if hasattr(lhs_value, inplace_fn_name):
            result = self.handle_call(node, getattr(lhs_value, inplace_fn_name), rhs_value)
            if not self.is_not_implemented(result):
                if result is not lhs_value:
                    raise ValueError("Inplace operation must return the same object")
                self.handle_assign(node.target, result)
                return
        if hasattr(lhs_value, regular_fn_name):
            result = self.handle_call(node, getattr(lhs_value, regular_fn_name), rhs_value)
            if not self.is_not_implemented(result):
                self.handle_assign(node.target, result)
                return
        if hasattr(rhs_value, right_fn_name) and type(lhs_value) is not type(rhs_value):
            result = self.handle_call(node, getattr(rhs_value, right_fn_name), lhs_value)
            if not self.is_not_implemented(result):
                self.handle_assign(node.target, result)
                return
        raise TypeError(
            f"unsupported operand type(s) for {op_to_symbol[type(node.op)]}=: "
            f"'{type(lhs_value).__name__}' and '{type(rhs_value).__name__}'"
        )

    def visit_AnnAssign(self, node):
        value = self.visit(node.value)
        self.handle_assign(node.target, value)

    def visit_For(self, node):
        from sonolus.script.internal.tuple_impl import TupleImpl

        iterable = self.visit(node.iter)
        if isinstance(iterable, TupleImpl):
            # Unroll the loop
            break_ctxs = []
            for value in iterable.value:
                set_ctx(ctx().branch(None))
                self.loop_head_ctxs.append([])
                self.break_ctxs.append([])
                self.handle_assign(node.target, validate_value(value))
                for stmt in node.body:
                    if not ctx().live:
                        break
                    self.visit(stmt)
                continue_ctxs = [*self.loop_head_ctxs.pop(), ctx()]
                break_ctxs.extend(self.break_ctxs.pop())
                set_ctx(Context.meet(continue_ctxs))
            if break_ctxs:
                set_ctx(Context.meet([*break_ctxs, ctx()]))
            return
        iterator = self.handle_call(node, iterable.__iter__)
        if not isinstance(iterator, SonolusIterator):
            raise ValueError("Unsupported iterator")
        writes = scan_writes(node)
        header_ctx = ctx().prepare_loop_header(writes)
        self.loop_head_ctxs.append(header_ctx)
        self.break_ctxs.append([])
        set_ctx(header_ctx)
        has_next = self.convert_to_boolean_num(node, self.handle_call(node, iterator.has_next))
        if has_next._is_py_() and not has_next._as_py_():
            # The loop will never run, continue after evaluating the condition
            self.loop_head_ctxs.pop()
            self.break_ctxs.pop()
            for stmt in node.orelse:
                if not ctx().live:
                    break
                self.visit(stmt)
            return
        ctx().test = has_next.ir()
        body_ctx = ctx().branch(None)
        else_ctx = ctx().branch(0)

        set_ctx(body_ctx)
        self.handle_assign(node.target, self.handle_call(node, iterator.next))
        for stmt in node.body:
            if not ctx().live:
                break
            self.visit(stmt)
        ctx().branch_to_loop_header(header_ctx)

        self.loop_head_ctxs.pop()
        break_ctxs = self.break_ctxs.pop()

        set_ctx(else_ctx)
        for stmt in node.orelse:
            if not ctx().live:
                break
            self.visit(stmt)
        else_end_ctx = ctx()

        after_ctx = Context.meet([else_end_ctx, *break_ctxs])
        set_ctx(after_ctx)

    def visit_While(self, node):
        writes = scan_writes(node)
        header_ctx = ctx().prepare_loop_header(writes)
        self.loop_head_ctxs.append(header_ctx)
        self.break_ctxs.append([])
        set_ctx(header_ctx)
        test = self.convert_to_boolean_num(node.test, self.visit(node.test))
        if test._is_py_():
            if test._as_py_():
                # The loop will run until a break / return
                body_ctx = ctx().branch(None)
                set_ctx(body_ctx)
                for stmt in node.body:
                    if not ctx().live:
                        break
                    self.visit(stmt)
                ctx().branch_to_loop_header(header_ctx)

                self.loop_head_ctxs.pop()
                break_ctxs = self.break_ctxs.pop()

                # Skip the else block

                after_ctx = Context.meet([ctx().into_dead(), *break_ctxs])
                set_ctx(after_ctx)
                return
            else:
                # The loop will never run, continue after evaluating the condition
                self.loop_head_ctxs.pop()
                self.break_ctxs.pop()
                for stmt in node.orelse:
                    if not ctx().live:
                        break
                    self.visit(stmt)
                return
        ctx().test = test.ir()
        body_ctx = ctx().branch(None)
        else_ctx = ctx().branch(0)

        set_ctx(body_ctx)
        for stmt in node.body:
            if not ctx().live:
                break
            self.visit(stmt)
        ctx().branch_to_loop_header(header_ctx)

        self.loop_head_ctxs.pop()
        break_ctxs = self.break_ctxs.pop()

        set_ctx(else_ctx)
        for stmt in node.orelse:
            if not ctx().live:
                break
            self.visit(stmt)
        else_end_ctx = ctx()

        after_ctx = Context.meet([else_end_ctx, *break_ctxs])
        set_ctx(after_ctx)

    def visit_If(self, node):
        test = self.convert_to_boolean_num(node.test, self.visit(node.test))

        if test._is_py_():
            if test._as_py_():
                for stmt in node.body:
                    if not ctx().live:
                        break
                    self.visit(stmt)
            else:
                for stmt in node.orelse:
                    if not ctx().live:
                        break
                    self.visit(stmt)
            return

        ctx_init = ctx()
        ctx_init.test = test.ir()
        true_ctx = ctx_init.branch(None)
        false_ctx = ctx_init.branch(0)

        set_ctx(true_ctx)
        for stmt in node.body:
            if not ctx().live:
                break
            self.visit(stmt)
        true_end_ctx = ctx()

        set_ctx(false_ctx)
        for stmt in node.orelse:
            if not ctx().live:
                break
            self.visit(stmt)
        false_end_ctx = ctx()

        set_ctx(Context.meet([true_end_ctx, false_end_ctx]))

    def visit_With(self, node):
        raise NotImplementedError("With statements are not supported")

    def visit_AsyncWith(self, node):
        raise NotImplementedError("Async with statements are not supported")

    def visit_Match(self, node):
        subject = self.visit(node.subject)
        end_ctxs = []
        for case in node.cases:
            if not ctx().live:
                break
            true_ctx, false_ctx = self.handle_match_pattern(subject, case.pattern)
            if not true_ctx.live:
                set_ctx(false_ctx)
                continue
            set_ctx(true_ctx)
            guard = (
                self.convert_to_boolean_num(case.guard, self.visit(case.guard)) if case.guard else validate_value(True)
            )
            if guard._is_py_():
                if guard._as_py_():
                    for stmt in case.body:
                        if not ctx().live:
                            break
                        self.visit(stmt)
                    end_ctxs.append(ctx())
                else:
                    # Merge failing before the guard and failing now at the guard (which we know is guaranteed to fail)
                    false_ctx = Context.meet([ctx(), false_ctx])
            else:
                ctx().test = guard.ir()
                guard_true_ctx = ctx().branch(None)
                guard_false_ctx = ctx().branch(0)
                set_ctx(guard_true_ctx)
                for stmt in case.body:
                    if not ctx().live:
                        break
                    self.visit(stmt)
                end_ctxs.append(ctx())
                false_ctx = Context.meet([false_ctx, guard_false_ctx])
            set_ctx(false_ctx)
        end_ctxs.append(ctx())
        if end_ctxs:
            set_ctx(Context.meet(end_ctxs))

    def handle_match_pattern(self, subject: Value, pattern: ast.pattern) -> tuple[Context, Context]:
        from sonolus.script.internal.generic import validate_type_spec
        from sonolus.script.internal.tuple_impl import TupleImpl

        if not ctx().live:
            return ctx().into_dead(), ctx()

        match pattern:
            case ast.MatchValue(value=value):
                value = self.visit(value)
                test = self.convert_to_boolean_num(pattern, validate_value(subject == value))
                if test._is_py_():
                    if test._as_py_():
                        return ctx(), ctx().into_dead()
                    else:
                        return ctx().into_dead(), ctx()
                ctx_init = ctx()
                ctx_init.test = test.ir()
                true_ctx = ctx_init.branch(None)
                false_ctx = ctx_init.branch(0)
                return true_ctx, false_ctx
            case ast.MatchSingleton(value=value):
                match value:
                    case True:
                        raise NotImplementedError("Matching against True is not supported, use 1 instead")
                    case False:
                        raise NotImplementedError("Matching against False is not supported, use 0 instead")
                    case None:
                        test = validate_value(subject._is_py_() and subject._as_py_() is None)
                    case _:
                        raise NotImplementedError("Unsupported match singleton")
                ctx_init = ctx()
                ctx_init.test = test.ir()
                true_ctx = ctx_init.branch(None)
                false_ctx = ctx_init.branch(0)
                return true_ctx, false_ctx
            case ast.MatchSequence(patterns=patterns):
                target_len = len(patterns)
                if not (isinstance(subject, Sequence | TupleImpl)):
                    return ctx().into_dead(), ctx()
                length_test = self.convert_to_boolean_num(pattern, validate_value(_len(subject) == target_len))
                ctx_init = ctx()
                if not length_test._is_py_():
                    ctx_init.test = length_test.ir()
                    true_ctx = ctx_init.branch(None)
                    false_ctxs = [ctx_init.branch(0)]
                elif length_test._as_py_():
                    true_ctx = ctx_init
                    false_ctxs = []
                else:
                    return ctx().into_dead(), ctx()
                set_ctx(true_ctx)
                for i, subpattern in enumerate(patterns):
                    if not ctx().live:
                        break
                    value = self.handle_getitem(subpattern, subject, validate_value(i))
                    true_ctx, false_ctx = self.handle_match_pattern(value, subpattern)
                    false_ctxs.append(false_ctx)
                    set_ctx(true_ctx)
                return true_ctx, Context.meet(false_ctxs)
            case ast.MatchMapping():
                raise NotImplementedError("Match mappings are not supported")
            case ast.MatchClass(cls=cls, patterns=patterns, kwd_attrs=kwd_attrs, kwd_patterns=kwd_patterns):
                cls = self.visit(cls)
                if cls._is_py_() and cls._as_py_() in {_int, _float, _bool}:
                    raise TypeError("Instance check against int, float, or bool is not supported, use Num instead")
                cls = validate_type_spec(cls)
                if not isinstance(cls, type):
                    raise TypeError("Class is not a type")
                if not isinstance(subject, cls):
                    return ctx().into_dead(), ctx()
                if patterns:
                    if not hasattr(cls, "__match_args__"):
                        raise TypeError("Class does not support match patterns")
                    if len(cls.__match_args__) < len(patterns):
                        raise ValueError("Too many match patterns")
                    # kwd_attrs can't be mixed with patterns on the syntax level,
                    # so we can just set it like this since it's empty
                    kwd_attrs = cls.__match_args__[: len(patterns)]
                    kwd_patterns = patterns
                if kwd_attrs:
                    true_ctx = ctx()
                    false_ctxs = []
                    for attr, subpattern in zip(kwd_attrs, kwd_patterns, strict=False):
                        if not hasattr(subject, attr):
                            raise AttributeError(f"Object has no attribute {attr}")
                        value = self.handle_getattr(subpattern, subject, attr)
                        true_ctx, false_ctx = self.handle_match_pattern(value, subpattern)
                        false_ctxs.append(false_ctx)
                        set_ctx(true_ctx)
                    return true_ctx, Context.meet(false_ctxs)
                return ctx(), ctx().into_dead()
            case ast.MatchStar():
                raise NotImplementedError("Match stars are not supported")
            case ast.MatchAs(pattern=pattern, name=name):
                if pattern:
                    true_ctx, false_ctx = self.handle_match_pattern(subject, pattern)
                    if name:
                        true_ctx.scope.set_value(name, subject)
                    return true_ctx, false_ctx
                else:
                    if name:
                        ctx().scope.set_value(name, subject)
                    return ctx(), ctx().into_dead()
            case ast.MatchOr():
                true_ctxs = []
                assert pattern.patterns
                for subpattern in pattern.patterns:
                    if not ctx().live:
                        break
                    true_ctx, false_ctx = self.handle_match_pattern(subject, subpattern)
                    true_ctxs.append(true_ctx)
                    set_ctx(false_ctx)
                return Context.meet(true_ctxs), ctx()

    def visit_Raise(self, node):
        raise NotImplementedError("Raise statements are not supported")

    def visit_Try(self, node):
        raise NotImplementedError("Try statements are not supported")

    def visit_TryStar(self, node):
        raise NotImplementedError("Try* statements are not supported")

    def visit_Assert(self, node):
        self.handle_call(
            node, assert_true, self.visit(node.test), self.visit(node.msg) if node.msg else validate_value(None)
        )

    def visit_Import(self, node):
        raise NotImplementedError("Import statements are not supported")

    def visit_ImportFrom(self, node):
        raise NotImplementedError("Import statements are not supported")

    def visit_Global(self, node):
        raise NotImplementedError("Global statements are not supported")

    def visit_Nonlocal(self, node):
        raise NotImplementedError("Nonlocal statements are not supported")

    def visit_Expr(self, node):
        return self.visit(node.value)

    def visit_Pass(self, node):
        pass

    def visit_Break(self, node):
        self.break_ctxs[-1].append(ctx())
        set_ctx(ctx().into_dead())

    def visit_Continue(self, node):
        loop_head = self.loop_head_ctxs[-1]
        if isinstance(loop_head, list):
            loop_head.append(ctx())
        else:
            ctx().branch_to_loop_header(loop_head)
        set_ctx(ctx().into_dead())

    def visit_BoolOp(self, node) -> Value:
        match node.op:
            case ast.And():
                handler = self.handle_and
            case ast.Or():
                handler = self.handle_or
            case _:
                raise NotImplementedError(f"Unsupported bool operator {op_to_symbol[type(node.op)]}")

        if not node.values:
            raise ValueError("Bool operator requires at least one operand")
        if len(node.values) == 1:
            return self.visit(node.values[0])
        initial, *rest = node.values
        return handler(self.visit(initial), ast.copy_location(ast.BoolOp(op=node.op, values=rest), node))

    def visit_NamedExpr(self, node):
        value = self.visit(node.value)
        self.handle_assign(node.target, value)
        return value

    def visit_BinOp(self, node):
        lhs = self.visit(node.left)
        rhs = self.visit(node.right)
        op = bin_ops[type(node.op)]
        if lhs._is_py_() and rhs._is_py_():
            lhs_py = lhs._as_py_()
            rhs_py = rhs._as_py_()
            if isinstance(lhs_py, type) and isinstance(rhs_py, type):
                return validate_value(getattr(lhs_py, op)(rhs_py))
        if hasattr(lhs, op):
            result = self.handle_call(node, getattr(lhs, op), rhs)
            if not self.is_not_implemented(result):
                return result
        if hasattr(rhs, rbin_ops[type(node.op)]) and type(lhs) is not type(rhs):
            result = self.handle_call(node, getattr(rhs, rbin_ops[type(node.op)]), lhs)
            if not self.is_not_implemented(result):
                return result
        raise TypeError(
            f"unsupported operand type(s) for {op_to_symbol[type(node.op)]}: "
            f"'{type(lhs).__name__}' and '{type(rhs).__name__}'"
        )

    def visit_UnaryOp(self, node):
        operand = self.visit(node.operand)
        if isinstance(node.op, ast.Not):
            return self.convert_to_boolean_num(node, operand).not_()
        op = unary_ops[type(node.op)]
        if operand._is_py_():
            operand_py = operand._as_py_()
            if isinstance(operand_py, type) and hasattr(type(operand_py), op):
                return self.handle_call(node, getattr(type(operand_py), op), operand_py)
        if hasattr(operand, op):
            return self.handle_call(node, getattr(operand, op))
        raise TypeError(f"bad operand type for unary {op_to_symbol[type(node.op)]}: '{type(operand).__name__}'")

    def visit_Lambda(self, node):
        signature = self.arguments_to_signature(node.args)

        def fn(*args, **kwargs):
            bound = signature.bind(*args, **kwargs)
            bound.apply_defaults()
            return Visitor(
                self.source_file,
                bound,
                self.globals,
                self,
            ).run(node)

        fn._meta_fn_ = True
        fn.__name__ = "<lambda>"

        return validate_value(fn)

    def visit_IfExp(self, node):
        test = self.convert_to_boolean_num(node.test, self.visit(node.test))

        if test._is_py_():
            if test._as_py_():
                return self.visit(node.body)
            else:
                return self.visit(node.orelse)

        res_name = self.new_name("ifexp")
        ctx_init = ctx()
        ctx_init.test = test.ir()

        set_ctx(ctx_init.branch(None))
        true_value = self.visit(node.body)
        ctx().scope.set_value(res_name, true_value)
        ctx_true = ctx()

        set_ctx(ctx_init.branch(0))
        false_value = self.visit(node.orelse)
        ctx().scope.set_value(res_name, false_value)
        ctx_false = ctx()

        set_ctx(Context.meet([ctx_true, ctx_false]))
        return ctx().scope.get_value(res_name)

    def visit_Dict(self, node):
        return validate_value({self.visit(k): self.visit(v) for k, v in zip(node.keys, node.values, strict=True)})

    def visit_Set(self, node):
        raise NotImplementedError("Set literals are not supported")

    def visit_ListComp(self, node):
        raise NotImplementedError("List comprehensions are not supported")

    def visit_SetComp(self, node):
        raise NotImplementedError("Set comprehensions are not supported")

    def visit_DictComp(self, node):
        raise NotImplementedError("Dict comprehensions are not supported")

    def visit_GeneratorExp(self, node):
        raise NotImplementedError("Generator expressions are not supported")

    def visit_Await(self, node):
        raise NotImplementedError("Await expressions are not supported")

    def visit_Yield(self, node):
        raise NotImplementedError("Yield expressions are not supported")

    def visit_YieldFrom(self, node):
        raise NotImplementedError("Yield from expressions are not supported")

    def _has_real_method(self, obj: Value, method_name: str) -> bool:
        return hasattr(obj, method_name) and not isinstance(getattr(obj, method_name), MethodWrapperType)

    def visit_Compare(self, node):
        result_name = self.new_name("compare")
        ctx().scope.set_value(result_name, Num._accept_(0))
        l_val = self.visit(node.left)
        false_ctxs = []
        for i, (op, rhs) in enumerate(zip(node.ops, node.comparators, strict=True)):
            r_val = self.visit(rhs)
            inverted = isinstance(op, ast.NotIn)
            result = None
            if isinstance(op, ast.Is | ast.IsNot):
                if not (r_val._is_py_() and r_val._as_py_() is None):
                    raise TypeError("The right operand of 'is' must be None")
                if isinstance(op, ast.Is):
                    result = Num._accept_(l_val._is_py_() and l_val._as_py_() is None)
                else:
                    result = Num._accept_(not (l_val._is_py_() and l_val._as_py_() is None))
            elif type(op) in comp_ops and self._has_real_method(l_val, comp_ops[type(op)]):
                result = self.handle_call(node, getattr(l_val, comp_ops[type(op)]), r_val)
            if (
                (result is None or self.is_not_implemented(result))
                and type(op) in rcomp_ops
                and self._has_real_method(r_val, rcomp_ops[type(op)])
            ):
                result = self.handle_call(node, getattr(r_val, rcomp_ops[type(op)]), l_val)
            if result is None or self.is_not_implemented(result):
                # Can't defer to the default object.__eq__ or similar since reference equality (is) is not reliable
                if type(op) is ast.Eq and type(l_val) is not type(r_val):
                    return Num._accept_(False)
                elif type(op) is ast.NotEq and type(l_val) is not type(r_val):
                    return Num._accept_(True)
                else:
                    raise TypeError(
                        f"'{op_to_symbol[type(op)]}' not supported between instances of '{type(l_val).__name__}' and "
                        f"'{type(r_val).__name__}'"
                    )
            result = self.ensure_boolean_num(result)
            if inverted:
                result = result.not_()
            curr_ctx = ctx()
            if i == len(node.ops) - 1:
                curr_ctx.scope.set_value(result_name, result)
            elif result._is_py_():
                if result._as_py_():
                    l_val = r_val
                else:
                    false_ctxs.append(curr_ctx)
                    set_ctx(curr_ctx.into_dead())
                    break
            else:
                curr_ctx.test = result.ir()
                true_ctx = curr_ctx.branch(None)
                false_ctx = curr_ctx.branch(0)
                false_ctxs.append(false_ctx)
                set_ctx(true_ctx)
                l_val = r_val
        last_ctx = ctx()  # This is the result of the last comparison returning true
        set_ctx(Context.meet([last_ctx, *false_ctxs]))
        return ctx().scope.get_value(result_name)

    def visit_Call(self, node):
        from sonolus.script.internal.dict_impl import DictImpl

        fn = self.visit(node.func)
        args = []
        kwargs = {}
        for arg in node.args:
            if isinstance(arg, ast.Starred):
                args.extend(self.handle_starred(self.visit(arg.value)))
            else:
                args.append(self.visit(arg))
        for keyword in node.keywords:
            if keyword.arg:
                kwargs[keyword.arg] = self.visit(keyword.value)
            else:
                value = self.visit(keyword.value)
                if isinstance(value, DictImpl):
                    if not all(isinstance(k, str) for k in value.value):
                        raise ValueError("Keyword arguments must be strings")
                    kwargs.update(value.value)
                else:
                    raise ValueError("Starred keyword arguments (**kwargs) must be dictionaries")
        return self.handle_call(node, fn, *args, **kwargs)

    def visit_FormattedValue(self, node):
        raise NotImplementedError("F-strings are not supported")

    def visit_JoinedStr(self, node):
        raise NotImplementedError("F-strings are not supported")

    def visit_Constant(self, node):
        return validate_value(node.value)

    def visit_Attribute(self, node):
        return self.handle_getattr(node, self.visit(node.value), node.attr)

    def visit_Subscript(self, node):
        value = self.visit(node.value)
        slice_value = self.visit(node.slice)
        return self.handle_getitem(node, value, slice_value)

    def visit_Starred(self, node):
        raise NotImplementedError("Starred expressions are not supported")

    def visit_Name(self, node):
        self.active_ctx = ctx()
        v = self
        while v:
            if not isinstance(v.active_ctx.scope.get_binding(node.id), EmptyBinding):
                return v.active_ctx.scope.get_value(node.id)
            v = v.parent
        if node.id in self.globals:
            value = self.globals[node.id]
            if value is ctx:
                raise ValueError("Unexpected use of ctx in non meta-function")
            return validate_value(BUILTIN_IMPLS.get(id(value), value))
        raise NameError(f"Name {node.id} is not defined")

    def visit_List(self, node):
        raise NotImplementedError("List literals are not supported")

    def visit_Tuple(self, node):
        values = []
        for elt in node.elts:
            if isinstance(elt, ast.Starred):
                values.extend(self.handle_starred(self.visit(elt.value)))
            else:
                values.append(self.visit(elt))
        return validate_value(tuple(values))

    def visit_Slice(self, node):
        raise NotImplementedError("Slices are not supported")

    def handle_assign(self, target: ast.stmt | ast.expr, value: Value):
        match target:
            case ast.Name(id=name):
                ctx().scope.set_value(name, value)
            case ast.Attribute(value=attr_value, attr=attr):
                attr_value = self.visit(attr_value)
                self.handle_setattr(target, attr_value, attr, value)
            case ast.Subscript(value=sub_value, slice=slice_expr):
                sub_value = self.visit(sub_value)
                slice_value = self.visit(slice_expr)
                self.handle_setitem(target, sub_value, slice_value, value)
            case ast.Tuple(elts=elts) | ast.List(elts=elts):
                values = self.handle_starred(value)
                if len(elts) != len(values):
                    raise ValueError("Unpacking assignment requires the same number of elements")
                for elt, v in zip(elts, values, strict=False):
                    self.handle_assign(elt, validate_value(v))
            case ast.Starred():
                raise NotImplementedError("Starred assignment is not supported")
            case _:
                raise NotImplementedError("Unsupported assignment target")

    def handle_and(self, l_val: Value, r_expr: ast.expr) -> Value:
        ctx_init = ctx()
        l_val = self.ensure_boolean_num(l_val)

        if l_val._is_py_():
            if l_val._as_py_():
                # The rhs is definitely evaluated, so we can return it directly
                return self.ensure_boolean_num(self.visit(r_expr))
            else:
                return l_val

        ctx_init.test = l_val.ir()
        res_name = self.new_name("and")

        set_ctx(ctx_init.branch(None))
        r_val = self.ensure_boolean_num(self.visit(r_expr))
        ctx().scope.set_value(res_name, r_val)
        ctx_true = ctx()

        set_ctx(ctx_init.branch(0))
        ctx().scope.set_value(res_name, Num._accept_(0))
        ctx_false = ctx()

        set_ctx(Context.meet([ctx_true, ctx_false]))
        if l_val._is_py_() and r_val._is_py_():
            return Num._accept_(l_val._as_py_() and r_val._as_py_())
        return ctx().scope.get_value(res_name)

    def handle_or(self, l_val: Value, r_expr: ast.expr) -> Value:
        ctx_init = ctx()
        l_val = self.ensure_boolean_num(l_val)

        if l_val._is_py_():
            if l_val._as_py_():
                return l_val
            else:
                # The rhs is definitely evaluated, so we can return it directly
                return self.ensure_boolean_num(self.visit(r_expr))

        ctx_init.test = l_val.ir()
        res_name = self.new_name("or")

        set_ctx(ctx_init.branch(None))
        ctx().scope.set_value(res_name, l_val)
        ctx_true = ctx()

        set_ctx(ctx_init.branch(0))
        r_val = self.ensure_boolean_num(self.visit(r_expr))
        ctx().scope.set_value(res_name, r_val)
        ctx_false = ctx()

        set_ctx(Context.meet([ctx_true, ctx_false]))
        if l_val._is_py_() and r_val._is_py_():
            return Num._accept_(l_val._as_py_() or r_val._as_py_())
        return ctx().scope.get_value(res_name)

    def generic_visit(self, node):
        if isinstance(node, ast.stmt | ast.expr):
            with self.reporting_errors_at_node(node):
                raise NotImplementedError(f"Unsupported syntax: {type(node).__name__}")
        raise NotImplementedError(f"Unsupported syntax: {type(node).__name__}")

    def handle_getattr(self, node: ast.stmt | ast.expr, target: Value, key: str) -> Value:
        with self.reporting_errors_at_node(node):
            if isinstance(target, ConstantValue):
                # Unwrap so we can access fields
                target = target._as_py_()
            descriptor = None
            for cls in type.mro(type(target)):
                descriptor = cls.__dict__.get(key, None)
                if descriptor is not None:
                    break
            match descriptor:
                case property(fget=getter):
                    return self.handle_call(node, getter, target)
                case SonolusDescriptor() | FunctionType() | classmethod() | staticmethod() | None:
                    return validate_value(getattr(target, key))
                case non_descriptor if not hasattr(non_descriptor, "__get__"):
                    return validate_value(getattr(target, key))
                case _:
                    raise TypeError(f"Unsupported field or descriptor {key}")

    def handle_setattr(self, node: ast.stmt | ast.expr, target: Value, key: str, value: Value):
        with self.reporting_errors_at_node(node):
            if target._is_py_():
                target = target._as_py_()
            descriptor = getattr(type(target), key, None)
            match descriptor:
                case property(fset=setter):
                    if setter is None:
                        raise AttributeError(f"Cannot set attribute {key} because property has no setter")
                    self.handle_call(node, setter, target, value)
                case SonolusDescriptor():
                    setattr(target, key, value)
                case _:
                    raise TypeError(f"Unsupported field or descriptor {key}")

    def handle_call[**P, R](
        self, node: ast.stmt | ast.expr, fn: Callable[P, R], /, *args: P.args, **kwargs: P.kwargs
    ) -> R | Value:
        """Handles a call to the given callable."""
        self.active_ctx = ctx()
        if (
            isinstance(fn, Value)
            and fn._is_py_()
            and isinstance(fn._as_py_(), type)
            and issubclass(fn._as_py_(), Value)
        ):
            return validate_value(self.execute_at_node(node, fn._as_py_(), *args, **kwargs))
        else:
            return self.execute_at_node(node, lambda: validate_value(compile_and_call(fn, *args, **kwargs)))

    def handle_getitem(self, node: ast.stmt | ast.expr, target: Value, key: Value) -> Value:
        with self.reporting_errors_at_node(node):
            if target._is_py_() and isinstance(target._as_py_(), type):
                if not key._is_py_():
                    raise ValueError("Type parameters must be compile-time constants")
                return validate_value(target._as_py_()[key._as_py_()])
            else:
                if isinstance(target, Value) and hasattr(target, "__getitem__"):
                    return self.handle_call(node, target.__getitem__, key)
                raise TypeError(f"Cannot get items on {type(target).__name__}")

    def handle_setitem(self, node: ast.stmt | ast.expr, target: Value, key: Value, value: Value):
        with self.reporting_errors_at_node(node):
            if isinstance(target, Value) and hasattr(target, "__setitem__"):
                return self.handle_call(node, target.__setitem__, key, value)
            raise TypeError(f"Cannot set items on {type(target).__name__}")

    def handle_delitem(self, node: ast.stmt | ast.expr, target: Value, key: Value):
        with self.reporting_errors_at_node(node):
            if isinstance(target, Value) and hasattr(target, "__delitem__"):
                return self.handle_call(node, target.__delitem__, key)
            raise TypeError(f"Cannot delete items on {type(target).__name__}")

    def handle_starred(self, value: Value) -> tuple[Value, ...]:
        if isinstance(value, TupleImpl):
            return value.value
        raise ValueError("Unsupported starred expression")

    def is_not_implemented(self, value):
        value = validate_value(value)
        return value._is_py_() and value._as_py_() is NotImplemented

    def ensure_boolean_num(self, value) -> Num:
        # This just checks the type for now, although we could support custom __bool__ implementations in the future
        if not _is_num(value):
            raise TypeError(f"Invalid type where a bool (Num) was expected: {type(value).__name__}")
        return value

    def convert_to_boolean_num(self, node, value: Value) -> Num:
        if _is_num(value):
            return value
        if hasattr(type(value), "__bool__"):
            return self.ensure_boolean_num(self.handle_call(node, type(value).__bool__, validate_value(value)))
        if hasattr(type(value), "__len__"):
            length = self.handle_call(node, type(value).__len__, validate_value(value))
            if not _is_num(length):
                raise TypeError(f"Invalid type for __len__: {type(length).__name__}")
            if length._is_py_():
                return Num._accept_(length._as_py_() > 0)
            return length > Num._accept_(0)
        raise TypeError(f"Converting {type(value).__name__} to bool is not supported")

    def arguments_to_signature(self, arguments: ast.arguments) -> inspect.Signature:
        parameters: list[inspect.Parameter] = []
        pos_only_count = len(arguments.posonlyargs)
        for i, arg in enumerate(arguments.posonlyargs):
            default_idx = i - pos_only_count + len(arguments.defaults)
            default = self.visit(arguments.defaults[default_idx]) if default_idx >= 0 else None
            param = inspect.Parameter(
                name=arg.arg,
                kind=inspect.Parameter.POSITIONAL_ONLY,
                default=default if default_idx >= 0 else inspect.Parameter.empty,
                annotation=inspect.Parameter.empty,
            )
            parameters.append(param)

        pos_kw_count = len(arguments.args)
        for i, arg in enumerate(arguments.args):
            default_idx = i - pos_kw_count + len(arguments.defaults)
            default = self.visit(arguments.defaults[default_idx]) if default_idx >= 0 else None
            param = inspect.Parameter(
                name=arg.arg,
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                default=default if default_idx >= 0 else inspect.Parameter.empty,
                annotation=inspect.Parameter.empty,
            )
            parameters.append(param)

        if arguments.vararg:
            param = inspect.Parameter(
                name=arguments.vararg.arg,
                kind=inspect.Parameter.VAR_POSITIONAL,
                default=inspect.Parameter.empty,
                annotation=inspect.Parameter.empty,
            )
            parameters.append(param)

        for i, arg in enumerate(arguments.kwonlyargs):
            default = self.visit(arguments.kw_defaults[i]) if arguments.kw_defaults[i] is not None else None
            param = inspect.Parameter(
                name=arg.arg,
                kind=inspect.Parameter.KEYWORD_ONLY,
                default=default if default is not None else inspect.Parameter.empty,
                annotation=inspect.Parameter.empty,
            )
            parameters.append(param)

        if arguments.kwarg:
            param = inspect.Parameter(
                name=arguments.kwarg.arg,
                kind=inspect.Parameter.VAR_KEYWORD,
                default=inspect.Parameter.empty,
                annotation=inspect.Parameter.empty,
            )
            parameters.append(param)

        return inspect.Signature(parameters)

    def raise_exception_at_node(self, node: ast.stmt | ast.expr, cause: Exception) -> Never:
        """Throws a compilation error at the given node."""

        def thrower() -> Never:
            raise CompilationError(str(cause)) from cause

        self.execute_at_node(node, thrower)

    def execute_at_node[**P, R](
        self, node: ast.stmt | ast.expr, fn: Callable[P, R], /, *args: P.args, **kwargs: P.kwargs
    ) -> R:
        """Executes the given function at the given node for a better traceback."""
        location_args = {
            "lineno": node.lineno,
            "col_offset": node.col_offset,
            "end_lineno": node.end_lineno,
            "end_col_offset": node.end_col_offset,
        }

        expr = ast.Expression(
            body=ast.Call(
                func=ast.Name(id="fn", ctx=ast.Load(), **location_args),
                args=[
                    ast.Starred(
                        value=ast.Name(id="args", ctx=ast.Load(), **location_args),
                        ctx=ast.Load(),
                        **location_args,
                    )
                ],
                keywords=[
                    ast.keyword(
                        value=ast.Name(id="kwargs", ctx=ast.Load(), **location_args),
                        arg=None,
                        **location_args,
                    )
                ],
                **location_args,
            )
        )
        return eval(
            compile(expr, filename=self.source_file, mode="eval"),
            {"fn": fn, "args": args, "kwargs": kwargs, "_filter_traceback_": True},
        )

    def reporting_errors_at_node(self, node: ast.stmt | ast.expr):
        return ReportingErrorsAtNode(self, node)

    def new_name(self, name: str):
        self.used_names[name] = self.used_names.get(name, 0) + 1
        return f"${name}_{self.used_names[name]}"


# Not using @contextmanager so it doesn't end up in tracebacks
class ReportingErrorsAtNode:
    def __init__(self, compiler, node: ast.stmt | ast.expr):
        self.compiler = compiler
        self.node = node

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            return

        if issubclass(exc_type, CompilationError):
            raise exc_value from exc_value.__cause__

        if exc_value is not None:
            self.compiler.raise_exception_at_node(self.node, exc_value)



from __future__ import annotations

import random
from abc import abstractmethod
from collections.abc import Callable
from typing import Any

from sonolus.script.internal.context import ctx
from sonolus.script.internal.impl import meta_fn
from sonolus.script.iterator import SonolusIterator
from sonolus.script.num import Num
from sonolus.script.record import Record
from sonolus.script.values import copy

# Note: we don't use Range in this file because Range itself inherits from ArrayLike


class ArrayLike[T]:
    """Mixin for array-like objects.

    Inheritors must implement `__len__`, `__getitem__`, and `__setitem__`.

    Usage:
        ```python
        class MyArrayLike[T](Record, ArrayLike[T]):
            def __len__(self) -> int:
                ...

            def __getitem__(self, index: Num) -> T:
                ...

            def __setitem__(self, index: Num, value: T):
                ...
        ```
    """

    _allow_instance_check_ = True

    @abstractmethod
    def __len__(self) -> int:
        """Return the length of the array."""

    @abstractmethod
    def __getitem__(self, index: Num) -> T:
        """Return the item at the given index.

        Args:
            index: The index of the item. Must be an integer between 0 and `len(self) - 1`.
        """

    @abstractmethod
    def __setitem__(self, index: Num, value: T):
        """Set the value of the item at the given index.

        Args:
            index: The index of the item. Must be an integer between 0 and `len(self) - 1`.
            value: The value to set.
        """

    def __iter__(self) -> SonolusIterator[T]:
        """Return an iterator over the array."""
        return _ArrayIterator(0, self)

    def __contains__(self, value: Any) -> bool:
        """Return whether any element in the array is equal to the given value.

        Args:
            value: The value to check for.
        """
        i = 0
        while i < len(self):
            if self[i] == value:
                return True
            i += 1
        return False

    def __reversed__(self):
        """Return a reversed view of the array."""
        return _ArrayReverser(self)

    def _enumerate_(self, start: Num = 0) -> SonolusIterator[T]:
        return _ArrayEnumerator(0, start, self)

    def index(self, value: T, start: Num = 0, stop: Num | None = None) -> Num:
        """Return the index of the value in the array equal to the given value.

        Args:
            value: The value to search for.
            start: The index to start searching from.
            stop: The index to stop searching at. If `None`, search to the end of the array.
        """
        if stop is None:
            stop = len(self)
        i = start
        while i < stop:
            if self[i] == value:
                return i
            i += 1
        return -1

    def count(self, value: T) -> Num:
        """Return the number of elements in the array equal to the given value.

        Args:
            value: The value to count.
        """
        count = 0
        i = 0
        while i < len(self):
            if self[i] == value:
                count += 1
            i += 1
        return count

    def last_index(self, value: T) -> Num:
        """Return the last index of the value in the array equal to the given value.

        Args:
            value: The value to search for.
        """
        i = len(self) - 1
        while i >= 0:
            if self[i] == value:
                return i
            i -= 1
        return -1

    def index_of_max(self, *, key: Callable[T, Any] | None = None) -> Num:
        """Return the index of the maximum value in the array.

        Args:
            key: A one-argument ordering function to use for comparison like the one used in `max()`.
        """
        if len(self) == 0:
            return -1
        if key is None:
            key = _identity
        max_index = 0
        i = 1
        while i < len(self):
            if key(self[i]) > key(self[max_index]):
                max_index = i
            i += 1
        return max_index

    def index_of_min(self, *, key: Callable[T, Any] | None = None) -> Num:
        """Return the index of the minimum value in the array.

        Args:
            key: A one-argument ordering function to use for comparison like the one used in `min()`.
        """
        if len(self) == 0:
            return -1
        if key is None:
            key = _identity
        min_index = 0
        i = 1
        while i < len(self):
            if key(self[i]) < key(self[min_index]):
                min_index = i
            i += 1
        return min_index

    def _max_(self, key: Callable[T, Any] | None = None) -> T:
        index = self.index_of_max(key=key)
        assert index != -1
        return self[index]

    def _min_(self, key: Callable[T, Any] | None = None) -> T:
        index = self.index_of_min(key=key)
        assert index != -1
        return self[index]

    def swap(self, i: Num, j: Num, /):
        """Swap the values at the given indices.

        Args:
            i: The first index.
            j: The second index.
        """
        temp = copy(self[i])
        self[i] = self[j]
        self[j] = temp

    def sort(self, *, key: Callable[T, Any] | None = None, reverse: bool = False):
        """Sort the values in the array in place.

        Args:
            key: A one-argument ordering function to use for comparison.
            reverse: If `True`, sort in descending order, otherwise sort in ascending order.
        """
        if len(self) < 15 or key is not None:
            if key is None:
                key = _identity
            # May be worth adding a block sort variant for better performance on large arrays in the future
            _insertion_sort(self, 0, len(self), key, reverse)
        else:
            # Heap sort is unstable, so if there's a key, we can't rely on it
            _heap_sort(self, 0, len(self), reverse)

    def shuffle(self):
        """Shuffle the values in the array in place."""
        random.shuffle(self)  # type: ignore

    def reverse(self):
        """Reverse the values in the array in place."""
        i = 0
        j = len(self) - 1
        while i < j:
            self.swap(i, j)
            i += 1
            j -= 1


def _identity[T](value: T) -> T:
    return value


def _insertion_sort[T](array: ArrayLike[T], start: Num, end: Num, key: Callable[T, Any], reverse: bool):
    i = start + 1
    if reverse:
        while i < end:
            j = i
            while j > start and key(array[j - 1]) < key(array[j]):
                array.swap(j - 1, j)
                j -= 1
            i += 1
    else:
        while i < end:
            j = i
            while j > start and key(array[j - 1]) > key(array[j]):
                array.swap(j - 1, j)
                j -= 1
            i += 1


def _heapify[T](array: ArrayLike[T], end: Num, index: Num, reverse: bool):
    while True:
        left = index * 2 + 1
        right = left + 1
        largest = index
        if left < end and (array[left] > array[largest]) != reverse:
            largest = left
        if right < end and (array[right] > array[largest]) != reverse:
            largest = right
        if largest == index:
            break
        array.swap(index, largest)
        index = largest


def _heap_sort[T](array: ArrayLike[T], start: Num, end: Num, reverse: bool):
    i = end // 2 - 1
    while i >= start:
        _heapify(array, end, i, reverse)
        i -= 1
    i = end - 1
    while i > start:
        array.swap(start, i)
        _heapify(array, i, start, reverse)
        i -= 1


class _ArrayIterator[V: ArrayLike](Record, SonolusIterator):
    i: int
    array: V

    def has_next(self) -> bool:
        return self.i < len(self.array)

    def get(self) -> V:
        return self.array[self.i]

    def advance(self):
        self.i += 1


class _ArrayReverser[V: ArrayLike](Record, ArrayLike):
    array: V

    def __len__(self) -> int:
        return len(self.array)

    def __getitem__(self, index: Num) -> V:
        return self.array[len(self) - 1 - index]

    def __setitem__(self, index: Num, value: V):
        self.array[len(self) - 1 - index] = value

    def reversed(self) -> ArrayLike[V]:
        return self.array


class _ArrayEnumerator[V: ArrayLike](Record, SonolusIterator):
    i: int
    offset: int
    array: V

    def has_next(self) -> bool:
        return self.i < len(self.array)

    def get(self) -> tuple[int, Any]:
        return self.i + self.offset, self.array[self.i]

    def advance(self):
        self.i += 1


@meta_fn
def get_positive_index(index: Num, length: Num) -> Num:
    """Get the positive index for the given index in the array of the given length.

    This is used to convert negative indixes relative to the end of the array to positive indices.

    Args:
        index: The index to convert.
        length: The length of the array.

    Returns:
        The positive index.
    """
    if not ctx():
        return index if index >= 0 else index + length
    index = Num._accept_(index)
    length = Num._accept_(length)
    if index._is_py_() and length._is_py_():
        return Num._accept_(index._as_py_() + length._as_py_() if index._as_py_() < 0 else index._as_py_())
    else:
        return index + (index < 0) * length



from __future__ import annotations

from abc import ABC, abstractmethod
from collections import deque
from collections.abc import Sequence

from sonolus.backend.optimize.flow import BasicBlock


class CompilerPass(ABC):
    def requires(self) -> set[CompilerPass]:
        return set()

    def preserves(self) -> set[CompilerPass] | None:
        return None

    def destroys(self) -> set[CompilerPass] | None:
        return None

    def applies(self) -> set[CompilerPass]:
        return {self}

    def exists_after(self, passes: set[CompilerPass]) -> set[CompilerPass]:
        preserved = self.preserves()
        destroyed = self.destroys()
        if destroyed is None and preserved is None:
            return self.applies()
        if preserved is not None:
            passes = {p for p in passes if p in preserved}
        if destroyed is not None:
            passes = {p for p in passes if p not in destroyed}
        return passes | self.applies()

    @abstractmethod
    def run(self, entry: BasicBlock) -> BasicBlock:
        pass


def run_passes(entry: BasicBlock, passes: Sequence[CompilerPass]) -> BasicBlock:
    active_passes = set()
    queue = deque(passes)
    while queue:
        if len(queue) > 99:
            raise RuntimeError("Likely unsatisfiable pass requirements")
        current_pass = queue.popleft()
        missing_requirements = current_pass.requires() - active_passes
        if missing_requirements:
            queue.appendleft(current_pass)
            queue.extendleft(missing_requirements)
            continue
        entry = current_pass.run(entry)
        active_passes = current_pass.exists_after(active_passes)
    return entry



from collections.abc import Iterable
from typing import Any, ClassVar, Self

from sonolus.backend.place import BlockPlace
from sonolus.script.internal.impl import meta_fn
from sonolus.script.internal.value import DataValue, Value


class _Missing:
    def __repr__(self) -> str:
        return "MISSING"

    def __bool__(self) -> bool:
        return False


_MISSING = _Missing()


class ConstantValue(Value):
    """Wraps a python constant value usable in Sonolus scripts."""

    _parameterized_: ClassVar[dict[Any, type[Self]]] = {}
    _value: ClassVar[Any] = _MISSING
    instance: ClassVar[Self | _Missing] = _MISSING

    def __new__(cls) -> Self:
        if cls.value() is _MISSING:
            raise TypeError(f"Class {cls.__name__} is not parameterized")
        return cls.instance

    @classmethod
    def value(cls):
        # We need this to avoid descriptors getting in the way
        return cls._value[0] if cls._value is not _MISSING else _MISSING

    @classmethod
    def of(cls, value: Any) -> Self:
        if value in cls._parameterized_:
            return cls._parameterized_[value]()

        parameterized = cls._get_parameterized(value)
        cls._parameterized_[value] = parameterized
        return parameterized()

    @classmethod
    def _get_parameterized(cls, parameter: Any) -> type[Self]:
        class Parameterized(cls):
            _value = (parameter,)

        Parameterized.__name__ = f"{parameter}"
        Parameterized.__qualname__ = f"{parameter}"
        Parameterized.__module__ = cls.__module__
        Parameterized.instance = object.__new__(Parameterized)
        return Parameterized

    @classmethod
    def _is_concrete_(cls) -> bool:
        return True

    @classmethod
    def _size_(cls) -> int:
        return 0

    @classmethod
    def _is_value_type_(cls) -> bool:
        return False

    @classmethod
    def _from_place_(cls, place: BlockPlace) -> Self:
        if cls.value() is _MISSING:
            raise TypeError(f"Class {cls.__name__} is not parameterized")
        return cls()

    @classmethod
    def _accepts_(cls, value: Any) -> bool:
        from sonolus.script.internal.impl import validate_value

        # We rely on validate_value to create the correct instance
        return isinstance(validate_value(value), cls)

    @classmethod
    def _accept_(cls, value: Any) -> Self:
        from sonolus.script.internal.impl import validate_value

        # We rely on validate_value to create the correct instance
        value = validate_value(value)
        if not isinstance(value, cls):
            raise ValueError(f"Value {value} is not of type {cls}")
        return value

    def _is_py_(self) -> bool:
        return True

    def _as_py_(self) -> Any:
        return self.value()

    @classmethod
    def _from_list_(cls, values: Iterable[DataValue]) -> Self:
        return cls()

    def _to_list_(self, level_refs: dict[Any, str] | None = None) -> list[DataValue | str]:
        return []

    @classmethod
    def _flat_keys_(cls, prefix: str) -> list[str]:
        return []

    def _get_(self) -> Self:
        return self

    def _set_(self, value: Any):
        if value is not self:
            raise ValueError(f"{type(self).__name__} is immutable")

    def _copy_from_(self, value: Self):
        if value is not self:
            raise ValueError(f"{type(self).__name__} is immutable")

    def _copy_(self) -> Self:
        return self

    @classmethod
    def _alloc_(cls) -> Self:
        return cls()

    @classmethod
    def _zero_(cls) -> Self:
        return cls()

    @meta_fn
    def __eq__(self, other):
        return self is other

    @meta_fn
    def __ne__(self, other):
        return self is not other

    @meta_fn
    def __hash__(self):
        return hash(self.value())


class BasicConstantValue(ConstantValue):
    """For constants without any special behavior."""



import builtins
import sys
from collections.abc import Callable
from types import MethodType, ModuleType
from typing import Any, ClassVar, Self

from sonolus.script.internal.impl import validate_value

_MISSING = object()


class SimulationContext:
    """Context manager to simulate additional Sonolus runtime features like level memory."""

    _active_context: ClassVar[Self | None] = None
    additional_replacements: dict[Any, Any]
    values: dict[Any, Any]
    _original_values: dict[str, dict[str, Any]]  # module_name: {var_name: original_value}
    _original_import: Callable[[str, Any, Any, Any, Any], Any] | None

    def __init__(self, *, additional_replacements: dict[Any, Any] | None = None):
        if SimulationContext._active_context is not None:
            raise RuntimeError("SimulationContext is already active")
        SimulationContext._active_context = self
        self.additional_replacements = additional_replacements or {}
        self.values = {}
        self._original_values = {}
        self._original_import = None

    def get_or_put_value(self, key: Any, factory: Callable[[], Any]) -> Any:
        if key not in self.values:
            self.values[key] = validate_value(factory())
        return self.values[key]._get_()._as_py_()

    def set_or_put_value(self, key: Any, factory: Callable[[], Any], value: Any):
        if key not in self.values:
            self.values[key] = validate_value(factory())
        existing_value = self.values[key]
        existing_type = type(existing_value)
        value = existing_type._accept_(validate_value(value))
        if existing_type._is_value_type_():
            existing_value._set_(value)
        else:
            existing_value._copy_from_(value)

    def _get_replacement(self, value: Any) -> Any:
        try:
            if value in self.additional_replacements:
                return self.additional_replacements[value]
        except TypeError:
            pass
        if hasattr(value, "_get_sim_replacement_") and isinstance(value._get_sim_replacement_, MethodType):
            return value._get_sim_replacement_()
        return _MISSING

    def _substitute_module_variables(self, module) -> None:
        if not isinstance(module, ModuleType):
            return

        module_name = module.__name__

        if module_name in self._original_values:
            return

        original_values = {}

        for var_name, var_value in list(module.__dict__.items()):
            replacement = self._get_replacement(var_value)
            if replacement is not _MISSING:
                original_values[var_name] = var_value
                setattr(module, var_name, replacement)

        if original_values:
            self._original_values[module_name] = original_values

    def _update_loaded_modules(self) -> None:
        loaded_modules = list(sys.modules.values())

        for module in loaded_modules:
            if module is not None:
                self._substitute_module_variables(module)

    def _create_import_hook(self):
        original_import = builtins.__import__

        def hooked_import(name, globals=None, locals=None, fromlist=(), level=0):  # noqa: A002
            module = original_import(name, globals, locals, fromlist, level)
            if isinstance(module, ModuleType):
                self._substitute_module_variables(module)

            if fromlist:
                for item in fromlist:
                    if hasattr(module, item):
                        attr = getattr(module, item)
                        if isinstance(attr, ModuleType):
                            self._substitute_module_variables(attr)

            return module

        return hooked_import

    def __enter__(self) -> Self:
        self._update_loaded_modules()

        self._original_import = builtins.__import__
        builtins.__import__ = self._create_import_hook()

        return self

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any):
        try:
            if self._original_import is not None:
                builtins.__import__ = self._original_import
                self._original_import = None

            for module_name, original_values in self._original_values.items():
                if module_name in sys.modules:
                    module = sys.modules[module_name]
                    if isinstance(module, ModuleType):
                        for var_name, original_value in original_values.items():
                            setattr(module, var_name, original_value)

            self._original_values.clear()

        finally:
            SimulationContext._active_context = None


def sim_ctx() -> SimulationContext | None:
    """Get the current simulation context, or None if not active."""
    return SimulationContext._active_context



# ruff: noqa: A002
from dataclasses import dataclass
from typing import Annotated, Any, NewType, dataclass_transform, get_origin

from sonolus.backend.mode import Mode
from sonolus.backend.place import BlockPlace
from sonolus.script.debug import assert_unreachable
from sonolus.script.internal.context import ctx, debug_config
from sonolus.script.internal.descriptor import SonolusDescriptor
from sonolus.script.internal.generic import validate_concrete_type
from sonolus.script.internal.introspection import get_field_specifiers
from sonolus.script.internal.simulation_context import sim_ctx
from sonolus.script.num import Num
from sonolus.script.values import copy


@dataclass
class _SliderOption:
    name: str | None
    description: str | None
    standard: bool
    advanced: bool
    scope: str | None
    default: float
    min: float
    max: float
    step: float
    unit: str | None

    def to_dict(self):
        result = {
            "type": "slider",
            "name": self.name,
            "standard": self.standard,
            "advanced": self.advanced,
            "def": self.default,
            "min": self.min,
            "max": self.max,
            "step": self.step,
        }
        if self.description is not None:
            result["description"] = self.description
        if self.scope is not None:
            result["scope"] = self.scope
        if self.unit is not None:
            result["unit"] = self.unit
        return result


@dataclass
class _ToggleOption:
    name: str | None
    description: str | None
    standard: bool
    advanced: bool
    scope: str | None
    default: bool

    def to_dict(self):
        result = {
            "type": "toggle",
            "name": self.name,
            "standard": self.standard,
            "advanced": self.advanced,
            "def": int(self.default),
        }
        if self.description is not None:
            result["description"] = self.description
        if self.scope is not None:
            result["scope"] = self.scope
        return result


@dataclass
class _SelectOption:
    name: str | None
    description: str | None
    standard: bool
    advanced: bool
    scope: str | None
    default: int
    values: list[str]

    def to_dict(self):
        result = {
            "type": "select",
            "name": self.name,
            "standard": self.standard,
            "advanced": self.advanced,
            "def": self.default,
            "values": self.values,
        }
        if self.description is not None:
            result["description"] = self.description
        if self.scope is not None:
            result["scope"] = self.scope
        return result


def slider_option(
    *,
    name: str | None = None,
    description: str | None = None,
    standard: bool = False,
    advanced: bool = False,
    default: float,
    min: float,
    max: float,
    step: float,
    unit: str | None = None,
    scope: str | None = None,
) -> Any:
    """Define a slider option.

    Args:
        name: The name of the option.
        description: The description of the option.
        standard: Whether the option is standard.
        advanced: Whether the option is advanced.
        default: The default value of the option.
        min: The minimum value of the option.
        max: The maximum value of the option.
        step: The step value of the option.
        unit: The unit of the option.
        scope: The scope of the option.
    """
    return _SliderOption(name, description, standard, advanced, scope, default, min, max, step, unit)


def toggle_option(
    *,
    name: str | None = None,
    description: str | None = None,
    standard: bool = False,
    advanced: bool = False,
    default: bool,
    scope: str | None = None,
) -> Any:
    """Define a toggle option.

    Args:
        name: The name of the option.
        description: The description of the option.
        standard: Whether the option is standard.
        advanced: Whether the option is advanced.
        default: The default value of the option.
        scope: The scope of the option.
    """
    return _ToggleOption(name, description, standard, advanced, scope, default)


def select_option(
    *,
    name: str | None = None,
    description: str | None = None,
    standard: bool = False,
    advanced: bool = False,
    default: str | int,
    values: list[str],
    scope: str | None = None,
) -> Any:
    """Define a select option.

    Args:
        name: The name of the option.
        description: The description of the option.
        standard: Whether the option is standard.
        advanced: Whether the option is advanced.
        default: The default value of the option.
        values: The values of the option.
        scope: The scope of the option.
    """
    if isinstance(default, str):
        default = values.index(default)
    return _SelectOption(name, description, standard, advanced, scope, default, values)


type Options = NewType("Options", Any)
type _OptionInfo = _SliderOption | _ToggleOption | _SelectOption


class _OptionField(SonolusDescriptor):
    info: _OptionInfo
    index: int

    def __init__(self, info: _OptionInfo, index: int):
        self.info = info
        self.index = index

    def __get__(self, instance, owner):
        if sim_ctx():
            return sim_ctx().get_or_put_value((instance, self), lambda: copy(self.info.default))
        if ctx():
            match ctx().global_state.mode:
                case Mode.PLAY:
                    block = ctx().blocks.LevelOption
                case Mode.WATCH:
                    block = ctx().blocks.LevelOption
                case Mode.PREVIEW:
                    block = ctx().blocks.PreviewOption
                case Mode.TUTORIAL:
                    block = None
                case _:
                    assert_unreachable()
            if block is not None:
                return Num._from_place_(BlockPlace(block, self.index))
            else:
                return Num._accept_(self.info.default)
        raise RuntimeError("Options can only be accessed in a context")

    def __set__(self, instance, value):
        if sim_ctx():
            return sim_ctx().set_or_put_value((instance, self), lambda: copy(self.info.default), value)
        if ctx() and debug_config().unchecked_writes:
            match ctx().global_state.mode:
                case Mode.PLAY:
                    block = ctx().blocks.LevelOption
                case Mode.WATCH:
                    block = ctx().blocks.LevelOption
                case Mode.PREVIEW:
                    block = ctx().blocks.PreviewOption
                case Mode.TUTORIAL:
                    block = None
                case _:
                    assert_unreachable()
            if block is not None:
                Num._from_place_(BlockPlace(block, self.index))._set_(Num._accept_(value))
            else:
                raise RuntimeError("Options in the current mode cannot be set and use the default value")
        raise AttributeError("Options are read-only")


@dataclass_transform()
def options[T](cls: type[T]) -> T | Options:
    """Decorator to define options.

    Usage:
        ```python
        @options
        class Options:
            slider_option: float = slider_option(
                name='Slider Option',
                standard=True,
                advanced=False,
                default=0.5,
                min=0,
                max=1,
                step=0.1,
                unit='unit',
                scope='scope',
            )
            toggle_option: bool = toggle_option(
                name='Toggle Option',
                standard=True,
                advanced=False,
                default=True,
                scope='scope',
            )
            select_option: int = select_option(
                name='Select Option',
                standard=True,
                advanced=False,
                default='value',
                values=['value'],
                scope='scope',
            )
        ```
    """
    if len(cls.__bases__) != 1:
        raise ValueError("Options class must not inherit from any class (except object)")
    instance = cls()
    entries = []
    for i, (name, annotation) in enumerate(get_field_specifiers(cls).items()):
        if get_origin(annotation) is not Annotated:
            raise TypeError(f"Invalid annotation for options: {annotation}")
        annotation_type = annotation.__args__[0]
        annotation_values = annotation.__metadata__
        if len(annotation_values) != 1:
            raise ValueError("Invalid annotation values for options")
        annotation_type = validate_concrete_type(annotation_type)
        if annotation_type is not Num:
            raise TypeError(f"Invalid annotation type for options: {annotation_type}")
        annotation_value = annotation_values[0]
        if not isinstance(annotation_value, _SliderOption | _ToggleOption | _SelectOption):
            raise TypeError(f"Invalid annotation value for options: {annotation_value}")
        if annotation_value.name is None:
            annotation_value.name = name
        entries.append(annotation_value)
        setattr(cls, name, _OptionField(annotation_value, i))
    instance._options_ = entries
    instance._is_comptime_value_ = True
    return instance


@options
class EmptyOptions:
    pass



import gzip
import json
import struct
import sys
from collections.abc import Callable
from concurrent.futures import Executor
from concurrent.futures.thread import ThreadPoolExecutor
from dataclasses import dataclass
from os import process_cpu_count
from pathlib import Path

from sonolus.backend.mode import Mode
from sonolus.build.compile import compile_mode
from sonolus.script.archetype import _BaseArchetype
from sonolus.script.bucket import Buckets
from sonolus.script.effect import Effects
from sonolus.script.engine import EngineData, empty_play_mode, empty_preview_mode, empty_tutorial_mode, empty_watch_mode
from sonolus.script.instruction import (
    TutorialInstructionIcons,
    TutorialInstructions,
)
from sonolus.script.internal.callbacks import (
    navigate_callback,
    preprocess_callback,
    update_callback,
    update_spawn_callback,
)
from sonolus.script.internal.context import ReadOnlyMemory
from sonolus.script.options import Options
from sonolus.script.particle import Particles
from sonolus.script.project import BuildConfig
from sonolus.script.sprite import Skin
from sonolus.script.ui import UiConfig

type JsonValue = bool | int | float | str | list[JsonValue] | dict[str, JsonValue] | None


@dataclass
class PackagedEngine:
    configuration: bytes
    play_data: bytes
    watch_data: bytes
    preview_data: bytes
    tutorial_data: bytes
    rom: bytes

    def write(self, path: Path):
        path.mkdir(parents=True, exist_ok=True)
        (path / "EngineConfiguration").write_bytes(self.configuration)
        (path / "EnginePlayData").write_bytes(self.play_data)
        (path / "EngineWatchData").write_bytes(self.watch_data)
        (path / "EnginePreviewData").write_bytes(self.preview_data)
        (path / "EngineTutorialData").write_bytes(self.tutorial_data)
        (path / "EngineRom").write_bytes(self.rom)

    def read(self, path: Path):
        self.configuration = (path / "EngineConfiguration").read_bytes()
        self.play_data = (path / "EnginePlayData").read_bytes()
        self.watch_data = (path / "EngineWatchData").read_bytes()
        self.preview_data = (path / "EnginePreviewData").read_bytes()
        self.tutorial_data = (path / "EngineTutorialData").read_bytes()
        self.rom = (path / "EngineRom").read_bytes()


def no_gil() -> bool:
    return sys.version_info >= (3, 13) and not sys._is_gil_enabled()


def package_engine(
    engine: EngineData,
    config: BuildConfig | None = None,
):
    config = config or BuildConfig()
    rom = ReadOnlyMemory()
    configuration = build_engine_configuration(engine.options, engine.ui)
    if no_gil():
        thread_pool = ThreadPoolExecutor(process_cpu_count() or 1)
    else:
        thread_pool = None

    play_mode = engine.play if config.build_play else empty_play_mode()
    watch_mode = engine.watch if config.build_watch else empty_watch_mode()
    preview_mode = engine.preview if config.build_preview else empty_preview_mode()
    tutorial_mode = engine.tutorial if config.build_tutorial else empty_tutorial_mode()

    if thread_pool is not None:
        futures = {
            "play": thread_pool.submit(
                build_play_mode,
                archetypes=play_mode.archetypes,
                skin=play_mode.skin,
                effects=play_mode.effects,
                particles=play_mode.particles,
                buckets=play_mode.buckets,
                rom=rom,
                config=config,
                thread_pool=thread_pool,
            ),
            "watch": thread_pool.submit(
                build_watch_mode,
                archetypes=watch_mode.archetypes,
                skin=watch_mode.skin,
                effects=watch_mode.effects,
                particles=watch_mode.particles,
                buckets=watch_mode.buckets,
                rom=rom,
                update_spawn=watch_mode.update_spawn,
                config=config,
                thread_pool=thread_pool,
            ),
            "preview": thread_pool.submit(
                build_preview_mode,
                archetypes=preview_mode.archetypes,
                skin=preview_mode.skin,
                rom=rom,
                config=config,
                thread_pool=thread_pool,
            ),
            "tutorial": thread_pool.submit(
                build_tutorial_mode,
                skin=tutorial_mode.skin,
                effects=tutorial_mode.effects,
                particles=tutorial_mode.particles,
                instructions=tutorial_mode.instructions,
                instruction_icons=tutorial_mode.instruction_icons,
                preprocess=tutorial_mode.preprocess,
                navigate=tutorial_mode.navigate,
                update=tutorial_mode.update,
                rom=rom,
                config=config,
                thread_pool=thread_pool,
            ),
        }

        play_data = futures["play"].result()
        watch_data = futures["watch"].result()
        preview_data = futures["preview"].result()
        tutorial_data = futures["tutorial"].result()
    else:
        play_data = build_play_mode(
            archetypes=play_mode.archetypes,
            skin=play_mode.skin,
            effects=play_mode.effects,
            particles=play_mode.particles,
            buckets=play_mode.buckets,
            rom=rom,
            config=config,
            thread_pool=None,
        )
        watch_data = build_watch_mode(
            archetypes=watch_mode.archetypes,
            skin=watch_mode.skin,
            effects=watch_mode.effects,
            particles=watch_mode.particles,
            buckets=watch_mode.buckets,
            rom=rom,
            update_spawn=watch_mode.update_spawn,
            config=config,
            thread_pool=None,
        )
        preview_data = build_preview_mode(
            archetypes=preview_mode.archetypes,
            skin=preview_mode.skin,
            rom=rom,
            config=config,
            thread_pool=None,
        )
        tutorial_data = build_tutorial_mode(
            skin=tutorial_mode.skin,
            effects=tutorial_mode.effects,
            particles=tutorial_mode.particles,
            instructions=tutorial_mode.instructions,
            instruction_icons=tutorial_mode.instruction_icons,
            preprocess=tutorial_mode.preprocess,
            navigate=tutorial_mode.navigate,
            update=tutorial_mode.update,
            rom=rom,
            config=config,
            thread_pool=None,
        )

    return PackagedEngine(
        configuration=package_output(configuration),
        play_data=package_output(play_data),
        watch_data=package_output(watch_data),
        preview_data=package_output(preview_data),
        tutorial_data=package_output(tutorial_data),
        rom=package_rom(rom),
    )


def build_engine_configuration(
    options: Options,
    ui: UiConfig,
) -> JsonValue:
    return {
        "options": [option.to_dict() for option in options._options_],
        "ui": ui.to_dict(),
    }


def build_play_mode(
    archetypes: list[type[_BaseArchetype]],
    skin: Skin,
    effects: Effects,
    particles: Particles,
    buckets: Buckets,
    rom: ReadOnlyMemory,
    config: BuildConfig,
    thread_pool: Executor | None = None,
):
    return {
        **compile_mode(
            mode=Mode.PLAY,
            rom=rom,
            archetypes=archetypes,
            global_callbacks=None,
            passes=config.passes,
            thread_pool=thread_pool,
        ),
        "skin": build_skin(skin),
        "effect": build_effects(effects),
        "particle": build_particles(particles),
        "buckets": build_buckets(buckets),
    }


def build_watch_mode(
    archetypes: list[type[_BaseArchetype]],
    skin: Skin,
    effects: Effects,
    particles: Particles,
    buckets: Buckets,
    rom: ReadOnlyMemory,
    update_spawn: Callable[[], float],
    config: BuildConfig,
    thread_pool: Executor | None = None,
):
    return {
        **compile_mode(
            mode=Mode.WATCH,
            rom=rom,
            archetypes=archetypes,
            global_callbacks=[(update_spawn_callback, update_spawn)],
            passes=config.passes,
            thread_pool=thread_pool,
        ),
        "skin": build_skin(skin),
        "effect": build_effects(effects),
        "particle": build_particles(particles),
        "buckets": build_buckets(buckets),
    }


def build_preview_mode(
    archetypes: list[type[_BaseArchetype]],
    skin: Skin,
    rom: ReadOnlyMemory,
    config: BuildConfig,
    thread_pool: Executor | None = None,
):
    return {
        **compile_mode(
            mode=Mode.PREVIEW,
            rom=rom,
            archetypes=archetypes,
            global_callbacks=None,
            passes=config.passes,
            thread_pool=thread_pool,
        ),
        "skin": build_skin(skin),
    }


def build_tutorial_mode(
    skin: Skin,
    effects: Effects,
    particles: Particles,
    instructions: TutorialInstructions,
    instruction_icons: TutorialInstructionIcons,
    preprocess: Callable[[], None],
    navigate: Callable[[], None],
    update: Callable[[], None],
    rom: ReadOnlyMemory,
    config: BuildConfig,
    thread_pool: Executor | None = None,
):
    return {
        **compile_mode(
            mode=Mode.TUTORIAL,
            rom=rom,
            archetypes=[],
            global_callbacks=[
                (preprocess_callback, preprocess),
                (navigate_callback, navigate),
                (update_callback, update),
            ],
            passes=config.passes,
            thread_pool=thread_pool,
        ),
        "skin": build_skin(skin),
        "effect": build_effects(effects),
        "particle": build_particles(particles),
        "instruction": build_instructions(instructions, instruction_icons),
    }


def build_skin(skin: Skin) -> JsonValue:
    return {
        "renderMode": skin.render_mode,
        "sprites": [{"name": name, "id": i} for i, name in enumerate(skin._sprites_)],
    }


def build_effects(effects: Effects) -> JsonValue:
    return {"clips": [{"name": name, "id": i} for i, name in enumerate(effects._effects_)]}


def build_particles(particles: Particles) -> JsonValue:
    return {"effects": [{"name": name, "id": i} for i, name in enumerate(particles._particles_)]}


def build_buckets(buckets: Buckets) -> JsonValue:
    return [bucket.to_dict() for bucket in buckets._buckets_]


def build_instructions(instructions: TutorialInstructions, instruction_icons: TutorialInstructionIcons) -> JsonValue:
    return {
        "texts": [{"name": name, "id": i} for i, name in enumerate(instructions._instructions_)],
        "icons": [{"name": name, "id": i} for i, name in enumerate(instruction_icons._instruction_icons_)],
    }


def package_rom(rom: ReadOnlyMemory) -> bytes:
    values = rom.values or [0]
    output = bytearray()

    for value in values:
        output.extend(struct.pack("<f", value))

    return gzip.compress(bytes(output))


def package_output(value: JsonValue) -> bytes:
    json_data = json.dumps(value, separators=(",", ":")).encode("utf-8")
    return gzip.compress(json_data)



from math import pi

from hypothesis import strategies as st

from sonolus.script.vec import Vec2
from tests.script.conftest import is_close, run_and_validate

floats = st.floats(min_value=-999, max_value=999, allow_nan=False, allow_infinity=False)
nonzero_floats = floats.filter(lambda x: abs(x) > 1e-2)


def test_magnitude():
    def fn():
        v = Vec2(3, 4)
        return v.magnitude

    assert run_and_validate(fn) == 5


def test_angle():
    def fn():
        v = Vec2(1, 1)
        return v.angle

    assert is_close(run_and_validate(fn), 0.25 * pi)


def test_dot():
    def fn():
        v = Vec2(1, 2)
        u = Vec2(3, 4)
        return v.dot(u)

    assert run_and_validate(fn) == 11


def test_rotate():
    def fn():
        v = Vec2(1, 1)
        return v.rotate(0.25 * pi)

    result = run_and_validate(fn)
    assert is_close(result.x, 0)
    assert is_close(result.y, 2**0.5)


def test_rotate_about():
    def fn():
        v = Vec2(1, 1)
        pivot = Vec2(1, 0)
        return v.rotate_about(0.5 * pi, pivot)

    assert is_close(run_and_validate(fn).magnitude, 0)


def test_add():
    def fn():
        v = Vec2(1, 2)
        u = Vec2(3, 4)
        return v + u

    assert run_and_validate(fn) == Vec2(4, 6)


def test_sub():
    def fn():
        v = Vec2(1, 2)
        u = Vec2(3, 4)
        return v - u

    assert run_and_validate(fn) == Vec2(-2, -2)


def test_mul_vec():
    def fn():
        v = Vec2(1, 2)
        u = Vec2(3, 4)
        return v * u

    assert run_and_validate(fn) == Vec2(3, 8)


def test_mul_float():
    def fn():
        v = Vec2(1, 2)
        return v * 3

    assert run_and_validate(fn) == Vec2(3, 6)


def test_div_vec():
    def fn():
        v = Vec2(1, 2)
        u = Vec2(3, 4)
        return v / u

    assert run_and_validate(fn) == Vec2(1 / 3, 2 / 4)


def test_div_float():
    def fn():
        v = Vec2(1, 2)
        return v / 3

    assert run_and_validate(fn) == Vec2(1 / 3, 2 / 3)


def test_neg():
    def fn():
        v = Vec2(1, 2)
        return -v

    assert run_and_validate(fn) == Vec2(-1, -2)


def test_equal():
    def fn():
        v = Vec2(1, 2)
        u = Vec2(1, 2)
        return v == u

    assert run_and_validate(fn)


def test_not_equal():
    def fn():
        v = Vec2(1, 2)
        u = Vec2(3, 4)
        return v != u

    assert run_and_validate(fn)



# ruff: noqa: PLW1641, PT017
import pytest
from hypothesis import assume, given
from hypothesis import strategies as st

from sonolus.script.debug import debug_log
from sonolus.script.internal.error import CompilationError
from sonolus.script.record import Record
from tests.script.conftest import run_and_validate, run_compiled


class DefaultsOnly(Record):
    pass


class AddOnly(Record):
    def __add__(self, other):
        debug_log(1)
        return AddOnly()


class RAddOnly(Record):
    def __radd__(self, other):
        debug_log(2)
        return RAddOnly()


class IAddOnly(Record):
    def __iadd__(self, other):
        debug_log(3)
        return self


class AddAndRAdd(Record):
    def __add__(self, other):
        debug_log(4)
        return AddAndRAdd()

    def __radd__(self, other):
        debug_log(5)
        return AddAndRAdd()


class AddAndIAdd(Record):
    def __add__(self, other):
        debug_log(6)
        return AddAndIAdd()

    def __iadd__(self, other):
        debug_log(7)
        return self


class RAddAndIAdd(Record):
    def __radd__(self, other):
        debug_log(8)
        return RAddAndIAdd()

    def __iadd__(self, other):
        debug_log(9)
        return self


class AllAddOps(Record):
    def __add__(self, other):
        debug_log(10)
        return AllAddOps()

    def __radd__(self, other):
        debug_log(11)
        return AllAddOps()

    def __iadd__(self, other):
        debug_log(12)
        return self


class AllAddNotImplemented(Record):
    def __add__(self, other):
        debug_log(13)
        return NotImplemented

    def __radd__(self, other):
        debug_log(14)
        return NotImplemented

    def __iadd__(self, other):
        debug_log(15)
        return NotImplemented


class EqOnly(Record):
    def __eq__(self, other):
        debug_log(16)
        return True


class EqNotImplemented(Record):
    def __eq__(self, other):
        debug_log(17)
        return NotImplemented

    def __ne__(self, other):
        debug_log(18)
        return NotImplemented


class LtOnly(Record):
    def __lt__(self, other):
        debug_log(19)
        return True


class GtOnly(Record):
    def __gt__(self, other):
        debug_log(20)
        return True


class LtGt(Record):
    def __lt__(self, other):
        debug_log(21)
        return True

    def __gt__(self, other):
        debug_log(22)
        return True


class LtNotImplemented(Record):
    def __lt__(self, other):
        debug_log(23)
        return NotImplemented


class GtNotImplemented(Record):
    def __gt__(self, other):
        debug_log(24)
        return NotImplemented


class LtGtNotImplemented(Record):
    def __lt__(self, other):
        debug_log(25)
        return NotImplemented

    def __gt__(self, other):
        debug_log(26)
        return NotImplemented


class HasCall(Record):
    def __call__(self):
        debug_log(27)
        return 123


class BoolTrue(Record):
    def __bool__(self):
        debug_log(28)
        return True


class BoolFalse(Record):
    def __bool__(self):
        debug_log(29)
        return False


bin_values = [
    AllAddOps(),
    AllAddNotImplemented(),
    AddOnly(),
    RAddOnly(),
    IAddOnly(),
    AddAndRAdd(),
    AddAndIAdd(),
    RAddAndIAdd(),
    DefaultsOnly(),
]

eq_values = [
    EqOnly(),
    EqNotImplemented(),
    DefaultsOnly(),
]

comp_values = [
    LtGt(),
    LtGtNotImplemented(),
    LtOnly(),
    GtOnly(),
    LtNotImplemented(),
    GtNotImplemented(),
    DefaultsOnly(),
]


@given(
    st.one_of(*[st.just(value) for value in bin_values]),
    st.one_of(*[st.just(value) for value in bin_values]),
)
def test_bin_op(left, right):
    def fn():
        return left + right

    try:
        run_and_validate(fn)
    except TypeError as e:
        assert "unsupported operand type(s)" in str(e)


@given(
    st.one_of(*[st.just(value) for value in bin_values]),
    st.one_of(*[st.just(value) for value in bin_values]),
)
def test_iop(left, right):
    def fn():
        x = left
        y = right
        x += y

    try:
        run_and_validate(fn)
    except TypeError as e:
        assert "unsupported operand type(s)" in str(e)


@given(
    st.one_of(*[st.just(value) for value in eq_values]),
    st.one_of(*[st.just(value) for value in eq_values]),
)
def test_eq_op(left, right):
    assume(not (isinstance(left, EqNotImplemented) and isinstance(right, EqNotImplemented)))  # See the next test

    def fn():
        return left == right

    run_and_validate(fn)


def test_eq_not_implemented():
    def fn():
        x = EqNotImplemented()
        y = EqNotImplemented()
        return x == y

    assert not fn()
    with pytest.raises(CompilationError, match="not supported between instances"):
        run_compiled(fn)


def test_not_eq_not_implemented():
    def fn():
        x = EqNotImplemented()
        y = EqNotImplemented()
        return x != y

    assert fn()
    with pytest.raises(CompilationError, match="not supported between instances"):
        run_compiled(fn)


@given(
    st.one_of(*[st.just(value) for value in comp_values]),
    st.one_of(*[st.just(value) for value in comp_values]),
)
def test_comp_op(left, right):
    def fn():
        return left < right

    try:
        run_and_validate(fn)
    except TypeError as e:
        assert "not supported between instances" in str(e)


def test_call_op():
    def fn():
        x = HasCall()
        return x()

    run_and_validate(fn)


def test_unsupported_call():
    def fn():
        x = DefaultsOnly()
        return x()

    try:
        run_and_validate(fn)
    except TypeError as e:
        assert "not callable" in str(e)


def test_unsupported_unary():
    def fn():
        x = DefaultsOnly()
        return -x  # type: ignore

    try:
        run_and_validate(fn)
    except TypeError as e:
        assert "bad operand type" in str(e)


def test_bool_true_truthiness():
    def fn():
        x = BoolTrue()
        return 1 if x else 0

    assert run_and_validate(fn) == 1


def test_bool_false_truthiness():
    def fn():
        x = BoolFalse()
        return 1 if x else 0

    assert run_and_validate(fn) == 0


def test_bool_true_match_case():
    def fn():
        x = BoolTrue()
        match x:
            case _ if x:
                return 1
            case _:
                return 0

    assert run_and_validate(fn) == 1


def test_bool_false_match_case():
    def fn():
        x = BoolFalse()
        match x:
            case _ if x:
                return 1
            case _:
                return 0

    assert run_and_validate(fn) == 0


def test_bool_true_while_condition():
    def fn():
        x = BoolTrue()
        while x:
            return 1
        return 0

    assert run_and_validate(fn) == 1


def test_bool_false_while_condition():
    def fn():
        x = BoolFalse()
        while x:
            return 1
        return 0

    assert run_and_validate(fn) == 0


def test_bool_true_not_operator():
    def fn():
        x = BoolTrue()
        return 1 if not x else 0

    assert run_and_validate(fn) == 0


def test_bool_false_not_operator():
    def fn():
        x = BoolFalse()
        return 1 if not x else 0

    assert run_and_validate(fn) == 1


def test_bool_call_true():
    def fn():
        x = BoolTrue()
        return bool(x)

    assert run_and_validate(fn)


def test_bool_call_false():
    def fn():
        x = BoolFalse()
        return bool(x)

    assert not run_and_validate(fn)



from collections.abc import Callable, Sequence
from contextvars import ContextVar
from typing import Any, Literal, Never

from sonolus.backend.mode import Mode
from sonolus.backend.ops import Op
from sonolus.backend.optimize.constant_evaluation import SparseConditionalConstantPropagation
from sonolus.backend.optimize.copy_coalesce import CopyCoalesce
from sonolus.backend.optimize.dead_code import (
    AdvancedDeadCodeElimination,
    DeadCodeElimination,
    UnreachableCodeElimination,
)
from sonolus.backend.optimize.flow import cfg_to_mermaid
from sonolus.backend.optimize.inlining import InlineVars
from sonolus.backend.optimize.passes import CompilerPass, run_passes
from sonolus.backend.optimize.simplify import CoalesceFlow, NormalizeSwitch, RewriteToSwitch
from sonolus.backend.optimize.ssa import FromSSA, ToSSA
from sonolus.script.internal.context import GlobalContextState, ReadOnlyMemory, ctx, set_ctx
from sonolus.script.internal.impl import meta_fn, validate_value
from sonolus.script.internal.native import native_function
from sonolus.script.internal.simulation_context import SimulationContext
from sonolus.script.num import Num

debug_log_callback = ContextVar[Callable[[Num], None]]("debug_log_callback")


@meta_fn
def error(message: str | None = None) -> Never:
    """Raise an error.

    This function is used to raise an error during runtime.
    When this happens, the game will pause in debug mode. The current callback will also immediately return 0.
    """
    message = validate_value(message)._as_py_() if message is not None else "Error"
    if not isinstance(message, str):
        raise ValueError("Expected a string")
    if ctx():
        debug_log(ctx().map_constant(message))
        debug_pause()
        terminate()
    else:
        raise RuntimeError(message)


@meta_fn
def static_error(message: str | None = None) -> Never:
    """Raise a static error.

    This function is used to raise an error during compile-time if the compiler cannot guarantee that
    this function will not be called during runtime.
    """
    message = validate_value(message)._as_py_() if message is not None else "Error"
    if not isinstance(message, str):
        raise ValueError("Expected a string")
    raise RuntimeError(message)


@meta_fn
def debug_log(value: Num):
    """Log a value in debug mode."""
    if debug_log_callback.get(None):
        return debug_log_callback.get()(value)
    else:
        return _debug_log(value)


@native_function(Op.DebugLog)
def _debug_log(value: Num):
    print(f"[DEBUG] {value}")
    return 0


@native_function(Op.DebugPause)
def debug_pause():
    """Pause the game if in debug mode."""
    input("[DEBUG] Paused")


def assert_true(value: Num, message: str | None = None):
    message = message if message is not None else "Assertion failed"
    if not value:
        error(message)


def assert_false(value: Num, message: str | None = None):
    message = message if message is not None else "Assertion failed"
    if value:
        error(message)


@meta_fn
def assert_unreachable(message: str | None = None) -> Never:
    message = validate_value(message)._as_py_() or "Unreachable code reached"
    raise RuntimeError(message)


@meta_fn
def terminate():
    if ctx():
        set_ctx(ctx().into_dead())
    else:
        raise RuntimeError("Terminated")


def visualize_cfg(
    fn: Callable[[], Any] | Callable[[type], Any],
    /,
    *,
    mode: Mode = Mode.PLAY,
    archetype: type | None = None,
    archetypes: list[type] | None,
    passes: Sequence[CompilerPass] | Literal["minimal", "basic", "standard"] = "basic",
) -> str:
    from sonolus.build.compile import callback_to_cfg

    match passes:
        case "minimal":
            passes = [
                CoalesceFlow(),
            ]
        case "basic":
            passes = [
                CoalesceFlow(),
                UnreachableCodeElimination(),
                AdvancedDeadCodeElimination(),
                CoalesceFlow(),
            ]
        case "standard":
            passes = [
                CoalesceFlow(),
                UnreachableCodeElimination(),
                DeadCodeElimination(),
                ToSSA(),
                SparseConditionalConstantPropagation(),
                UnreachableCodeElimination(),
                DeadCodeElimination(),
                CoalesceFlow(),
                InlineVars(),
                DeadCodeElimination(),
                RewriteToSwitch(),
                FromSSA(),
                CoalesceFlow(),
                CopyCoalesce(),
                AdvancedDeadCodeElimination(),
                CoalesceFlow(),
                NormalizeSwitch(),
            ]

    global_state = GlobalContextState(
        mode,
        {a: i for i, a in enumerate(archetypes)} if archetypes is not None else None,
        ReadOnlyMemory(),
    )

    cfg = callback_to_cfg(global_state, fn, "", archetype=archetype)
    cfg = run_passes(cfg, passes)
    return cfg_to_mermaid(cfg)


def simulation_context() -> SimulationContext:
    return SimulationContext()



import functools
import inspect
from collections.abc import Callable

from sonolus.backend.ir import IRInstr, IRPureInstr, IRSet
from sonolus.backend.ops import Op
from sonolus.script.internal.context import ctx
from sonolus.script.internal.impl import meta_fn, validate_value
from sonolus.script.num import Num, _is_num


def native_call(op: Op, *args: Num) -> Num:
    if not ctx():
        raise RuntimeError("Unexpected native call")
    args = tuple(validate_value(arg) for arg in args)
    if not all(_is_num(arg) for arg in args):
        raise RuntimeError("All arguments must be of type Num")
    result = ctx().alloc(size=1)
    ctx().add_statements(IRSet(result, (IRPureInstr if op.pure else IRInstr)(op, [arg.ir() for arg in args])))
    return Num._from_place_(result)


def native_function[**P, R](op: Op) -> Callable[[Callable[P, R]], Callable[P, R]]:
    def decorator(fn: Callable[P, Num]) -> Callable[P, Num]:
        signature = inspect.signature(fn)

        @functools.wraps(fn)
        @meta_fn
        def wrapper(*args: Num) -> Num:
            if len(args) < sum(1 for p in signature.parameters.values() if p.default == inspect.Parameter.empty):
                raise TypeError(f"Expected {len(signature.parameters)} arguments, got {len(args)}")
            if ctx():
                bound_args = signature.bind(*args)
                bound_args.apply_defaults()
                return native_call(op, *(Num._accept_(arg) for arg in bound_args.args))
            return fn(*args)

        return wrapper

    return decorator



from enum import StrEnum


class Op(StrEnum):
    def __new__(cls, name: str, side_effects: bool, pure: bool, control_flow: bool):
        obj = str.__new__(cls, name)
        obj._value_ = name
        obj.side_effects = side_effects
        obj.pure = pure
        obj.control_flow = control_flow
        return obj

    Abs = ("Abs", False, True, False)
    Add = ("Add", False, True, False)
    And = ("And", False, True, True)
    Arccos = ("Arccos", False, True, False)
    Arcsin = ("Arcsin", False, True, False)
    Arctan = ("Arctan", False, True, False)
    Arctan2 = ("Arctan2", False, True, False)
    BeatToBPM = ("BeatToBPM", False, True, False)
    BeatToStartingBeat = ("BeatToStartingBeat", False, True, False)
    BeatToStartingTime = ("BeatToStartingTime", False, True, False)
    BeatToTime = ("BeatToTime", False, True, False)
    Block = ("Block", False, True, True)
    Break = ("Break", True, False, True)
    Ceil = ("Ceil", False, True, False)
    Clamp = ("Clamp", False, True, False)
    Copy = ("Copy", True, False, False)
    Cos = ("Cos", False, True, False)
    Cosh = ("Cosh", False, True, False)
    DebugLog = ("DebugLog", True, False, False)
    DebugPause = ("DebugPause", True, False, False)
    DecrementPost = ("DecrementPost", True, False, False)
    DecrementPostPointed = ("DecrementPostPointed", True, False, False)
    DecrementPostShifted = ("DecrementPostShifted", True, False, False)
    DecrementPre = ("DecrementPre", True, False, False)
    DecrementPrePointed = ("DecrementPrePointed", True, False, False)
    DecrementPreShifted = ("DecrementPreShifted", True, False, False)
    Degree = ("Degree", False, True, False)
    DestroyParticleEffect = ("DestroyParticleEffect", True, False, False)
    Divide = ("Divide", False, True, False)
    DoWhile = ("DoWhile", False, True, True)
    Draw = ("Draw", True, False, False)
    DrawCurvedB = ("DrawCurvedB", True, False, False)
    DrawCurvedBT = ("DrawCurvedBT", True, False, False)
    DrawCurvedL = ("DrawCurvedL", True, False, False)
    DrawCurvedLR = ("DrawCurvedLR", True, False, False)
    DrawCurvedR = ("DrawCurvedR", True, False, False)
    DrawCurvedT = ("DrawCurvedT", True, False, False)
    EaseInBack = ("EaseInBack", False, True, False)
    EaseInCirc = ("EaseInCirc", False, True, False)
    EaseInCubic = ("EaseInCubic", False, True, False)
    EaseInElastic = ("EaseInElastic", False, True, False)
    EaseInExpo = ("EaseInExpo", False, True, False)
    EaseInOutBack = ("EaseInOutBack", False, True, False)
    EaseInOutCirc = ("EaseInOutCirc", False, True, False)
    EaseInOutCubic = ("EaseInOutCubic", False, True, False)
    EaseInOutElastic = ("EaseInOutElastic", False, True, False)
    EaseInOutExpo = ("EaseInOutExpo", False, True, False)
    EaseInOutQuad = ("EaseInOutQuad", False, True, False)
    EaseInOutQuart = ("EaseInOutQuart", False, True, False)
    EaseInOutQuint = ("EaseInOutQuint", False, True, False)
    EaseInOutSine = ("EaseInOutSine", False, True, False)
    EaseInQuad = ("EaseInQuad", False, True, False)
    EaseInQuart = ("EaseInQuart", False, True, False)
    EaseInQuint = ("EaseInQuint", False, True, False)
    EaseInSine = ("EaseInSine", False, True, False)
    EaseOutBack = ("EaseOutBack", False, True, False)
    EaseOutCirc = ("EaseOutCirc", False, True, False)
    EaseOutCubic = ("EaseOutCubic", False, True, False)
    EaseOutElastic = ("EaseOutElastic", False, True, False)
    EaseOutExpo = ("EaseOutExpo", False, True, False)
    EaseOutInBack = ("EaseOutInBack", False, True, False)
    EaseOutInCirc = ("EaseOutInCirc", False, True, False)
    EaseOutInCubic = ("EaseOutInCubic", False, True, False)
    EaseOutInElastic = ("EaseOutInElastic", False, True, False)
    EaseOutInExpo = ("EaseOutInExpo", False, True, False)
    EaseOutInQuad = ("EaseOutInQuad", False, True, False)
    EaseOutInQuart = ("EaseOutInQuart", False, True, False)
    EaseOutInQuint = ("EaseOutInQuint", False, True, False)
    EaseOutInSine = ("EaseOutInSine", False, True, False)
    EaseOutQuad = ("EaseOutQuad", False, True, False)
    EaseOutQuart = ("EaseOutQuart", False, True, False)
    EaseOutQuint = ("EaseOutQuint", False, True, False)
    EaseOutSine = ("EaseOutSine", False, True, False)
    Equal = ("Equal", False, True, False)
    Execute = ("Execute", False, True, False)
    Execute0 = ("Execute0", False, True, False)
    ExportValue = ("ExportValue", True, False, False)
    Floor = ("Floor", False, True, False)
    Frac = ("Frac", False, True, False)
    Get = ("Get", False, False, False)
    GetPointed = ("GetPointed", False, False, False)
    GetShifted = ("GetShifted", False, False, False)
    Greater = ("Greater", False, True, False)
    GreaterOr = ("GreaterOr", False, True, False)
    HasEffectClip = ("HasEffectClip", False, True, False)
    HasParticleEffect = ("HasParticleEffect", False, True, False)
    HasSkinSprite = ("HasSkinSprite", False, True, False)
    If = ("If", False, True, True)
    IncrementPost = ("IncrementPost", True, False, False)
    IncrementPostPointed = ("IncrementPostPointed", True, False, False)
    IncrementPostShifted = ("IncrementPostShifted", True, False, False)
    IncrementPre = ("IncrementPre", True, False, False)
    IncrementPrePointed = ("IncrementPrePointed", True, False, False)
    IncrementPreShifted = ("IncrementPreShifted", True, False, False)
    Judge = ("Judge", False, True, False)
    JudgeSimple = ("JudgeSimple", False, True, False)
    JumpLoop = ("JumpLoop", False, True, True)
    Lerp = ("Lerp", False, True, False)
    LerpClamped = ("LerpClamped", False, True, False)
    Less = ("Less", False, True, False)
    LessOr = ("LessOr", False, True, False)
    Log = ("Log", False, True, False)
    Max = ("Max", False, True, False)
    Min = ("Min", False, True, False)
    Mod = ("Mod", False, True, False)
    MoveParticleEffect = ("MoveParticleEffect", True, False, False)
    Multiply = ("Multiply", False, True, False)
    Negate = ("Negate", False, True, False)
    Not = ("Not", False, True, False)
    NotEqual = ("NotEqual", False, True, False)
    Or = ("Or", False, True, True)
    Paint = ("Paint", True, False, False)
    Play = ("Play", True, False, False)
    PlayLooped = ("PlayLooped", True, False, False)
    PlayLoopedScheduled = ("PlayLoopedScheduled", True, False, False)
    PlayScheduled = ("PlayScheduled", True, False, False)
    Power = ("Power", False, True, False)
    Print = ("Print", True, False, False)
    Radian = ("Radian", False, True, False)
    Random = ("Random", False, False, False)
    RandomInteger = ("RandomInteger", False, False, False)
    Rem = ("Rem", False, True, False)
    Remap = ("Remap", False, True, False)
    RemapClamped = ("RemapClamped", False, True, False)
    Round = ("Round", False, True, False)
    Set = ("Set", True, False, False)
    SetAdd = ("SetAdd", True, False, False)
    SetAddPointed = ("SetAddPointed", True, False, False)
    SetAddShifted = ("SetAddShifted", True, False, False)
    SetDivide = ("SetDivide", True, False, False)
    SetDividePointed = ("SetDividePointed", True, False, False)
    SetDivideShifted = ("SetDivideShifted", True, False, False)
    SetMod = ("SetMod", True, False, False)
    SetModPointed = ("SetModPointed", True, False, False)
    SetModShifted = ("SetModShifted", True, False, False)
    SetMultiply = ("SetMultiply", True, False, False)
    SetMultiplyPointed = ("SetMultiplyPointed", True, False, False)
    SetMultiplyShifted = ("SetMultiplyShifted", True, False, False)
    SetPointed = ("SetPointed", True, False, False)
    SetPower = ("SetPower", True, False, False)
    SetPowerPointed = ("SetPowerPointed", True, False, False)
    SetPowerShifted = ("SetPowerShifted", True, False, False)
    SetRem = ("SetRem", True, False, False)
    SetRemPointed = ("SetRemPointed", True, False, False)
    SetRemShifted = ("SetRemShifted", True, False, False)
    SetShifted = ("SetShifted", True, False, False)
    SetSubtract = ("SetSubtract", True, False, False)
    SetSubtractPointed = ("SetSubtractPointed", True, False, False)
    SetSubtractShifted = ("SetSubtractShifted", True, False, False)
    Sign = ("Sign", False, True, False)
    Sin = ("Sin", False, True, False)
    Sinh = ("Sinh", False, True, False)
    Spawn = ("Spawn", True, False, False)
    SpawnParticleEffect = ("SpawnParticleEffect", True, False, False)
    StackEnter = ("StackEnter", True, False, False)
    StackGet = ("StackGet", False, False, False)
    StackGetFrame = ("StackGetFrame", False, False, False)
    StackGetFramePointer = ("StackGetFramePointer", False, False, False)
    StackGetPointer = ("StackGetPointer", False, False, False)
    StackGrow = ("StackGrow", True, False, False)
    StackInit = ("StackInit", True, False, False)
    StackLeave = ("StackLeave", True, False, False)
    StackPop = ("StackPop", True, False, False)
    StackPush = ("StackPush", True, False, False)
    StackSet = ("StackSet", True, False, False)
    StackSetFrame = ("StackSetFrame", True, False, False)
    StackSetFramePointer = ("StackSetFramePointer", True, False, False)
    StackSetPointer = ("StackSetPointer", True, False, False)
    StopLooped = ("StopLooped", True, False, False)
    StopLoopedScheduled = ("StopLoopedScheduled", True, False, False)
    # Streams are immutable when they're readable, so we can treat read operations as pure.
    StreamGetNextKey = ("StreamGetNextKey", False, True, False)
    StreamGetPreviousKey = ("StreamGetPreviousKey", False, True, False)
    StreamGetValue = ("StreamGetValue", False, True, False)
    StreamHas = ("StreamHas", False, True, False)
    StreamSet = ("StreamSet", True, False, False)
    Subtract = ("Subtract", False, True, False)
    Switch = ("Switch", False, True, True)
    SwitchInteger = ("SwitchInteger", False, True, True)
    SwitchIntegerWithDefault = ("SwitchIntegerWithDefault", False, True, True)
    SwitchWithDefault = ("SwitchWithDefault", False, True, True)
    Tan = ("Tan", False, True, False)
    Tanh = ("Tanh", False, True, False)
    TimeToScaledTime = ("TimeToScaledTime", False, True, False)
    TimeToStartingScaledTime = ("TimeToStartingScaledTime", False, True, False)
    TimeToStartingTime = ("TimeToStartingTime", False, True, False)
    TimeToTimeScale = ("TimeToTimeScale", False, True, False)
    Trunc = ("Trunc", False, True, False)
    Unlerp = ("Unlerp", False, True, False)
    UnlerpClamped = ("UnlerpClamped", False, True, False)
    While = ("While", False, True, True)



from enum import Enum


class BlockData(int):
    def __new__(cls, id_: int, readable: set[str], writable: set[str]):
        obj = int.__new__(cls, id_)
        obj.readable = readable
        obj.writable = writable
        return obj


class BlockEnum(BlockData, Enum):
    def __str__(self):
        return self.name


class TutorialBlock(BlockEnum):
    RuntimeEnvironment = (1000, {"preprocess", "navigate", "update"}, {"preprocess"})
    RuntimeUpdate = (1001, {"preprocess", "navigate", "update"}, {})
    RuntimeSkinTransform = (1002, {"preprocess", "navigate", "update"}, {"preprocess", "navigate", "update"})
    RuntimeParticleTransform = (1003, {"preprocess", "navigate", "update"}, {"preprocess", "navigate", "update"})
    RuntimeBackground = (1004, {"preprocess", "navigate", "update"}, {"preprocess", "navigate", "update"})
    RuntimeUI = (1005, {"preprocess", "navigate", "update"}, {"preprocess"})
    RuntimeUIConfiguration = (1006, {"preprocess", "navigate", "update"}, {"preprocess"})
    TutorialMemory = (2000, {"preprocess", "navigate", "update"}, {"preprocess", "navigate", "update"})
    TutorialData = (2001, {"preprocess", "navigate", "update"}, {"preprocess"})
    TutorialInstruction = (2002, {"preprocess", "navigate", "update"}, {"preprocess", "navigate", "update"})
    EngineRom = (3000, {"preprocess", "navigate", "update"}, {})
    TemporaryMemory = (10000, {"preprocess", "navigate", "update"}, {"preprocess", "navigate", "update"})


class PlayBlock(BlockEnum):
    RuntimeEnvironment = (
        1000,
        {
            "preprocess",
            "spawnOrder",
            "shouldSpawn",
            "initialize",
            "updateSequential",
            "touch",
            "updateParallel",
            "terminate",
        },
        {"preprocess"},
    )
    RuntimeUpdate = (
        1001,
        {
            "preprocess",
            "spawnOrder",
            "shouldSpawn",
            "initialize",
            "updateSequential",
            "touch",
            "updateParallel",
            "terminate",
        },
        {},
    )
    RuntimeTouchArray = (
        1002,
        {
            "preprocess",
            "spawnOrder",
            "shouldSpawn",
            "initialize",
            "updateSequential",
            "touch",
            "updateParallel",
            "terminate",
        },
        {},
    )
    RuntimeSkinTransform = (
        1003,
        {
            "preprocess",
            "spawnOrder",
            "shouldSpawn",
            "initialize",
            "updateSequential",
            "touch",
            "updateParallel",
            "terminate",
        },
        {"preprocess", "updateSequential", "touch"},
    )
    RuntimeParticleTransform = (
        1004,
        {
            "preprocess",
            "spawnOrder",
            "shouldSpawn",
            "initialize",
            "updateSequential",
            "touch",
            "updateParallel",
            "terminate",
        },
        {"preprocess", "updateSequential", "touch"},
    )
    RuntimeBackground = (
        1005,
        {
            "preprocess",
            "spawnOrder",
            "shouldSpawn",
            "initialize",
            "updateSequential",
            "touch",
            "updateParallel",
            "terminate",
        },
        {"preprocess", "updateSequential", "touch"},
    )
    RuntimeUI = (
        1006,
        {
            "preprocess",
            "spawnOrder",
            "shouldSpawn",
            "initialize",
            "updateSequential",
            "touch",
            "updateParallel",
            "terminate",
        },
        {"preprocess"},
    )
    RuntimeUIConfiguration = (
        1007,
        {
            "preprocess",
            "spawnOrder",
            "shouldSpawn",
            "initialize",
            "updateSequential",
            "touch",
            "updateParallel",
            "terminate",
        },
        {"preprocess"},
    )
    LevelMemory = (
        2000,
        {
            "preprocess",
            "spawnOrder",
            "shouldSpawn",
            "initialize",
            "updateSequential",
            "touch",
            "updateParallel",
            "terminate",
        },
        {"preprocess", "updateSequential", "touch"},
    )
    LevelData = (
        2001,
        {
            "preprocess",
            "spawnOrder",
            "shouldSpawn",
            "initialize",
            "updateSequential",
            "touch",
            "updateParallel",
            "terminate",
        },
        {"preprocess"},
    )
    LevelOption = (
        2002,
        {
            "preprocess",
            "spawnOrder",
            "shouldSpawn",
            "initialize",
            "updateSequential",
            "touch",
            "updateParallel",
            "terminate",
        },
        {},
    )
    LevelBucket = (
        2003,
        {
            "preprocess",
            "spawnOrder",
            "shouldSpawn",
            "initialize",
            "updateSequential",
            "touch",
            "updateParallel",
            "terminate",
        },
        {"preprocess"},
    )
    LevelScore = (
        2004,
        {
            "preprocess",
            "spawnOrder",
            "shouldSpawn",
            "initialize",
            "updateSequential",
            "touch",
            "updateParallel",
            "terminate",
        },
        {"preprocess"},
    )
    LevelLife = (
        2005,
        {
            "preprocess",
            "spawnOrder",
            "shouldSpawn",
            "initialize",
            "updateSequential",
            "touch",
            "updateParallel",
            "terminate",
        },
        {"preprocess"},
    )
    EngineRom = (
        3000,
        {
            "preprocess",
            "spawnOrder",
            "shouldSpawn",
            "initialize",
            "updateSequential",
            "touch",
            "updateParallel",
            "terminate",
        },
        {},
    )
    EntityMemory = (
        4000,
        {
            "preprocess",
            "spawnOrder",
            "shouldSpawn",
            "initialize",
            "updateSequential",
            "touch",
            "updateParallel",
            "terminate",
        },
        {
            "preprocess",
            "spawnOrder",
            "shouldSpawn",
            "initialize",
            "updateSequential",
            "touch",
            "updateParallel",
            "terminate",
        },
    )
    EntityData = (
        4001,
        {
            "preprocess",
            "spawnOrder",
            "shouldSpawn",
            "initialize",
            "updateSequential",
            "touch",
            "updateParallel",
            "terminate",
        },
        {"preprocess"},
    )
    EntitySharedMemory = (
        4002,
        {
            "preprocess",
            "spawnOrder",
            "shouldSpawn",
            "initialize",
            "updateSequential",
            "touch",
            "updateParallel",
            "terminate",
        },
        {"preprocess", "updateSequential", "touch"},
    )
    EntityInfo = (
        4003,
        {
            "preprocess",
            "spawnOrder",
            "shouldSpawn",
            "initialize",
            "updateSequential",
            "touch",
            "updateParallel",
            "terminate",
        },
        {},
    )
    EntityDespawn = (
        4004,
        {
            "preprocess",
            "spawnOrder",
            "shouldSpawn",
            "initialize",
            "updateSequential",
            "touch",
            "updateParallel",
            "terminate",
        },
        {
            "preprocess",
            "spawnOrder",
            "shouldSpawn",
            "initialize",
            "updateSequential",
            "touch",
            "updateParallel",
            "terminate",
        },
    )
    EntityInput = (
        4005,
        {
            "preprocess",
            "spawnOrder",
            "shouldSpawn",
            "initialize",
            "updateSequential",
            "touch",
            "updateParallel",
            "terminate",
        },
        {
            "preprocess",
            "spawnOrder",
            "shouldSpawn",
            "initialize",
            "updateSequential",
            "touch",
            "updateParallel",
            "terminate",
        },
    )
    EntityDataArray = (
        4101,
        {
            "preprocess",
            "spawnOrder",
            "shouldSpawn",
            "initialize",
            "updateSequential",
            "touch",
            "updateParallel",
            "terminate",
        },
        {"preprocess"},
    )
    EntitySharedMemoryArray = (
        4102,
        {
            "preprocess",
            "spawnOrder",
            "shouldSpawn",
            "initialize",
            "updateSequential",
            "touch",
            "updateParallel",
            "terminate",
        },
        {"preprocess", "updateSequential", "touch"},
    )
    EntityInfoArray = (
        4103,
        {
            "preprocess",
            "spawnOrder",
            "shouldSpawn",
            "initialize",
            "updateSequential",
            "touch",
            "updateParallel",
            "terminate",
        },
        {},
    )
    ArchetypeLife = (
        5000,
        {
            "preprocess",
            "spawnOrder",
            "shouldSpawn",
            "initialize",
            "updateSequential",
            "touch",
            "updateParallel",
            "terminate",
        },
        {"preprocess"},
    )
    TemporaryMemory = (
        10000,
        {
            "preprocess",
            "spawnOrder",
            "shouldSpawn",
            "initialize",
            "updateSequential",
            "touch",
            "updateParallel",
            "terminate",
        },
        {
            "preprocess",
            "spawnOrder",
            "shouldSpawn",
            "initialize",
            "updateSequential",
            "touch",
            "updateParallel",
            "terminate",
        },
    )


class PreviewBlock(BlockEnum):
    RuntimeEnvironment = (1000, {"preprocess", "render"}, {"preprocess"})
    RuntimeCanvas = (1001, {"preprocess", "render"}, {"preprocess"})
    RuntimeSkinTransform = (1002, {"preprocess", "render"}, {"preprocess"})
    RuntimeUI = (1003, {"preprocess", "render"}, {"preprocess"})
    RuntimeUIConfiguration = (1004, {"preprocess", "render"}, {"preprocess"})
    PreviewData = (2000, {"preprocess", "render"}, {"preprocess"})
    PreviewOption = (2001, {"preprocess", "render"}, {})
    EngineRom = (3000, {"preprocess", "render"}, {})
    EntityData = (4000, {"preprocess", "render"}, {"preprocess"})
    EntitySharedMemory = (4001, {"preprocess", "render"}, {"preprocess"})
    EntityInfo = (4002, {"preprocess", "render"}, {})
    EntityDataArray = (4100, {"preprocess", "render"}, {"preprocess"})
    EntitySharedMemoryArray = (4101, {"preprocess", "render"}, {"preprocess"})
    EntityInfoArray = (4102, {"preprocess", "render"}, {})
    TemporaryMemory = (10000, {"preprocess", "render"}, {"preprocess", "render"})


class WatchBlock(BlockEnum):
    RuntimeEnvironment = (
        1000,
        {
            "preprocess",
            "spawnTime",
            "despawnTime",
            "initialize",
            "updateSequential",
            "updateParallel",
            "terminate",
            "updateSpawn",
        },
        {"preprocess"},
    )
    RuntimeUpdate = (
        1001,
        {
            "preprocess",
            "spawnTime",
            "despawnTime",
            "initialize",
            "updateSequential",
            "updateParallel",
            "terminate",
            "updateSpawn",
        },
        {},
    )
    RuntimeSkinTransform = (
        1002,
        {
            "preprocess",
            "spawnTime",
            "despawnTime",
            "initialize",
            "updateSequential",
            "updateParallel",
            "terminate",
            "updateSpawn",
        },
        {"preprocess", "updateSequential"},
    )
    RuntimeParticleTransform = (
        1003,
        {
            "preprocess",
            "spawnTime",
            "despawnTime",
            "initialize",
            "updateSequential",
            "updateParallel",
            "terminate",
            "updateSpawn",
        },
        {"preprocess", "updateSequential"},
    )
    RuntimeBackground = (
        1004,
        {
            "preprocess",
            "spawnTime",
            "despawnTime",
            "initialize",
            "updateSequential",
            "updateParallel",
            "terminate",
            "updateSpawn",
        },
        {"preprocess", "updateSequential"},
    )
    RuntimeUI = (
        1005,
        {
            "preprocess",
            "spawnTime",
            "despawnTime",
            "initialize",
            "updateSequential",
            "updateParallel",
            "terminate",
            "updateSpawn",
        },
        {"preprocess"},
    )
    RuntimeUIConfiguration = (
        1006,
        {
            "preprocess",
            "spawnTime",
            "despawnTime",
            "initialize",
            "updateSequential",
            "updateParallel",
            "terminate",
            "updateSpawn",
        },
        {"preprocess"},
    )
    LevelMemory = (
        2000,
        {
            "preprocess",
            "spawnTime",
            "despawnTime",
            "initialize",
            "updateSequential",
            "updateParallel",
            "terminate",
            "updateSpawn",
        },
        {"preprocess", "updateSequential"},
    )
    LevelData = (
        2001,
        {
            "preprocess",
            "spawnTime",
            "despawnTime",
            "initialize",
            "updateSequential",
            "updateParallel",
            "terminate",
            "updateSpawn",
        },
        {"preprocess"},
    )
    LevelOption = (
        2002,
        {
            "preprocess",
            "spawnTime",
            "despawnTime",
            "initialize",
            "updateSequential",
            "updateParallel",
            "terminate",
            "updateSpawn",
        },
        {},
    )
    LevelBucket = (
        2003,
        {
            "preprocess",
            "spawnTime",
            "despawnTime",
            "initialize",
            "updateSequential",
            "updateParallel",
            "terminate",
            "updateSpawn",
        },
        {"preprocess"},
    )
    LevelScore = (
        2004,
        {
            "preprocess",
            "spawnTime",
            "despawnTime",
            "initialize",
            "updateSequential",
            "updateParallel",
            "terminate",
            "updateSpawn",
        },
        {"preprocess"},
    )
    LevelLife = (
        2005,
        {
            "preprocess",
            "spawnTime",
            "despawnTime",
            "initialize",
            "updateSequential",
            "updateParallel",
            "terminate",
            "updateSpawn",
        },
        {"preprocess"},
    )
    EngineRom = (
        3000,
        {
            "preprocess",
            "spawnTime",
            "despawnTime",
            "initialize",
            "updateSequential",
            "updateParallel",
            "terminate",
            "updateSpawn",
        },
        {},
    )
    EntityMemory = (
        4000,
        {"preprocess", "spawnTime", "despawnTime", "initialize", "updateSequential", "updateParallel", "terminate"},
        {"preprocess", "spawnTime", "despawnTime", "initialize", "updateSequential", "updateParallel", "terminate"},
    )
    EntityData = (
        4001,
        {"preprocess", "spawnTime", "despawnTime", "initialize", "updateSequential", "updateParallel", "terminate"},
        {"preprocess"},
    )
    EntitySharedMemory = (
        4002,
        {"preprocess", "spawnTime", "despawnTime", "initialize", "updateSequential", "updateParallel", "terminate"},
        {"preprocess", "updateSequential"},
    )
    EntityInfo = (
        4003,
        {"preprocess", "spawnTime", "despawnTime", "initialize", "updateSequential", "updateParallel", "terminate"},
        {},
    )
    EntityInput = (
        4004,
        {"preprocess", "spawnTime", "despawnTime", "initialize", "updateSequential", "updateParallel", "terminate"},
        {"preprocess"},
    )
    EntityDataArray = (
        4101,
        {
            "preprocess",
            "spawnTime",
            "despawnTime",
            "initialize",
            "updateSequential",
            "updateParallel",
            "terminate",
            "updateSpawn",
        },
        {"preprocess"},
    )
    EntitySharedMemoryArray = (
        4102,
        {
            "preprocess",
            "spawnTime",
            "despawnTime",
            "initialize",
            "updateSequential",
            "updateParallel",
            "terminate",
            "updateSpawn",
        },
        {"preprocess", "updateSequential"},
    )
    EntityInfoArray = (
        4103,
        {
            "preprocess",
            "spawnTime",
            "despawnTime",
            "initialize",
            "updateSequential",
            "updateParallel",
            "terminate",
            "updateSpawn",
        },
        {},
    )
    ArchetypeLife = (
        5000,
        {
            "preprocess",
            "spawnTime",
            "despawnTime",
            "initialize",
            "updateSequential",
            "updateParallel",
            "terminate",
            "updateSpawn",
        },
        {"preprocess"},
    )
    TemporaryMemory = (
        10000,
        {
            "preprocess",
            "spawnTime",
            "despawnTime",
            "initialize",
            "updateSequential",
            "updateParallel",
            "terminate",
            "updateSpawn",
        },
        {
            "preprocess",
            "spawnTime",
            "despawnTime",
            "initialize",
            "updateSequential",
            "updateParallel",
            "terminate",
            "updateSpawn",
        },
    )


type Block = TutorialBlock | PlayBlock | PreviewBlock | WatchBlock



from threading import Lock
from typing import TypedDict

from sonolus.backend.node import ConstantNode, EngineNode, FunctionNode


class ValueOutputNode(TypedDict):
    value: float


class FunctionOutputNode(TypedDict):
    func: str
    args: list[int]


class OutputNodeGenerator:
    nodes: list[ValueOutputNode | FunctionOutputNode]
    indexes: dict[EngineNode, int]
    lock: Lock

    def __init__(self):
        self.nodes = []
        self.indexes = {}
        self.lock = Lock()

    def add(self, node: EngineNode):
        with self.lock:
            return self._add(node)

    def _add(self, node: EngineNode):
        if node in self.indexes:
            return self.indexes[node]

        match node:
            case ConstantNode(value):
                index = len(self.nodes)
                self.nodes.append({"value": value})
                self.indexes[node] = index
                return index
            case FunctionNode(func, args):
                arg_indexes = [self._add(arg) for arg in args]
                index = len(self.nodes)
                self.nodes.append({"func": func.value, "args": arg_indexes})
                self.indexes[node] = index
                return index
            case _:
                raise ValueError("Invalid node")

    def get(self):
        return self.nodes



# ruff: noqa
import builtins
from typing import (
    Any,
    Callable,
    Iterable,
    Iterator,
    Sequence,
    overload,
)

def all(iterable: Iterable[builtins.bool]) -> builtins.bool:
    """Return True if all elements of the iterable are true.

    Args:
        iterable: The iterable to evaluate.

    Returns:
        True if all elements are true, False otherwise.
    """
    ...

def any(iterable: Iterable[builtins.bool]) -> builtins.bool:
    """Return True if any element of the iterable is true.

    Args:
        iterable: The iterable to evaluate.

    Returns:
        True if any element is true, False otherwise.
    """
    ...

def abs(x: builtins.int | builtins.float) -> builtins.int | builtins.float:
    """Return the absolute value of a number.

    Args:
        x: A number.

    Returns:
        The absolute value of x.
    """
    ...

def bool(x: builtins.int | builtins.float | builtins.bool) -> builtins.bool:
    """Convert a value to a Boolean.

    Args:
        x: The value to convert.

    Returns:
        The Boolean value of x.
    """
    ...

def callable(obj: object) -> bool:
    """Check if the object appears callable.

    Args:
        obj: The object to check.

    Returns:
        True if the object appears callable, False otherwise.
    """
    ...

def enumerate[T](iterable: Iterable[T], start: int = 0) -> Iterator[tuple[int, T]]:
    """Return an enumerate object.

    Args:
        iterable: The iterable to enumerate.
        start: The starting index.

    Returns:
        An enumerate object.
    """
    ...

def filter[T](function: Callable[[T], builtins.bool] | None, iterable: Iterable[T]) -> Iterator[T]:
    """Construct an iterator from those elements of iterable for which function returns true.

    Args:
        function: A function that tests if each element should be included. If None, returns the elements that are true.
        iterable: The iterable to filter.

    Returns:
        An iterator yielding the filtered elements.
    """
    ...

def float(x: builtins.int | builtins.float) -> builtins.float:
    """Convert a number to a floating point number.

    Args:
        x: The number to convert.

    Returns:
        The floating point representation of x.
    """
    ...

def int(x: builtins.int | builtins.float) -> builtins.int:
    """Convert a number to an integer.

    Args:
        x: The number to convert.

    Returns:
        The integer representation of x.
    """
    ...

def isinstance(obj: object, classinfo: type | tuple[type, ...]) -> builtins.bool:
    """Check if an object is an instance of a class or of a subclass thereof.

    Args:
        obj: The object to check.
        classinfo: A type or a tuple of types.

    Returns:
        True if the object is an instance of classinfo, False otherwise.
    """
    ...

def issubclass(cls: type, classinfo: type | tuple[type, ...]) -> builtins.bool:
    """Check if a class is a subclass of another class or a tuple of classes.

    Args:
        cls: The class to check.
        classinfo: A class or a tuple of classes.

    Returns:
        True if cls is a subclass of classinfo, False otherwise.
    """
    ...

def len(s: object) -> builtins.int:
    """Return the number of items in a container.

    Args:
        s: The container object.

    Returns:
        The number of items in s.
    """
    ...

def map[T, S](function: Callable[[T], S], iterable: Iterable[T]) -> Iterator[S]:
    """Apply a function to every item of an iterable and return an iterator.

    Unlike the standard Python map function, it is possible that the function may be called more than once on the
    same item.

    Args:
        function: The function to apply.
        iterable: The iterable to process.

    Returns:
        An iterator with the results.
    """
    ...

@overload
def max[T](iterable: Iterable[T], *, key: Callable[[T], Any] | None = ...) -> T:
    """Return the largest item in an iterable or the largest of two or more arguments.

    Args:
        iterable: The iterable to evaluate.
        key: A function of one argument that is used to extract a comparison key from each element.

    Returns:
        The largest item.
    """
    ...

@overload
def max[T](arg1: T, arg2: T, *args: T, key: Callable[[T], Any] | None = ...) -> T:
    """Return the largest item in an iterable or the largest of two or more arguments.

    Args:
        arg1: First argument.
        arg2: Second argument.
        *args: Additional arguments.
        key: A function of one argument that is used to extract a comparison key from each element.

    Returns:
        The largest item.
    """
    ...

@overload
def min[T](iterable: Iterable[T], *, key: Callable[[T], Any] | None = ...) -> T:
    """Return the smallest item in an iterable or the smallest of two or more arguments.

    Args:
        iterable: The iterable to evaluate.
        key: A function of one argument that is used to extract a comparison key from each element.

    Returns:
        The smallest item.
    """
    ...

@overload
def min[T](arg1: T, arg2: T, *args: T, key: Callable[[T], Any] | None = ...) -> T:
    """Return the smallest item in an iterable or the smallest of two or more arguments.

    Args:
        arg1: First argument.
        arg2: Second argument.
        *args: Additional arguments.
        key: A function of one argument that is used to extract a comparison key from each element.

    Returns:
        The smallest item.
    """
    ...

@overload
def range(stop: builtins.int) -> builtins.range:
    """Return an immutable sequence of numbers from 0 to stop.

    Args:
        stop: Stop value.

    Returns:
        The range object.
    """
    ...

@overload
def range(start: builtins.int, stop: builtins.int, step: builtins.int = ...) -> builtins.range:
    """Return an immutable sequence of numbers from start to stop by step.

    Args:
        start: Start value.
        stop: Stop value.
        step: Step value.

    Returns:
        The range object.
    """
    ...

def reversed[T](seq: Sequence[T]) -> Iterator[T]:
    """Return a reverse iterator.

    Args:
        seq: The sequence to reverse.

    Returns:
        An iterator over the reversed sequence.
    """
    ...

def round(number: builtins.int | builtins.float, ndigits: builtins.int = ...) -> builtins.float:
    """Round a number to a given precision in decimal digits.

    Args:
        number: The number to round.
        ndigits: The number of decimal digits to round to.

    Returns:
        The rounded number.
    """
    ...

def zip[T](*iterables: Iterable[T]) -> Iterator[tuple[T, ...]]:
    """Return an iterator of tuples, where the i-th tuple contains the i-th element from each of the argument sequences.

    Args:
        *iterables: Iterables to aggregate.

    Returns:
        An iterator of aggregated tuples.
    """
    ...



import pytest
from hypothesis import given
from hypothesis import strategies as st

from sonolus.script.array import Array
from sonolus.script.containers import VarArray
from sonolus.script.debug import assert_false, assert_true
from sonolus.script.record import Record
from tests.script.conftest import run_and_validate
from tests.script.test_record import Simple


def test_array_constructor():
    def fn():
        return Array(1, 2, 3)

    assert list(run_and_validate(fn)) == [1, 2, 3]


@given(args=st.lists(st.integers(min_value=-9999, max_value=9999), min_size=1, max_size=10))
def test_array_spread(args):
    tuple_args = tuple(args)  # lists are not supported

    def fn():
        return Array(*tuple_args)

    assert list(run_and_validate(fn)) == list(args)


def test_array_constructor_with_type():
    def fn():
        return Array[int, 3](1, 2, 3)

    assert list(run_and_validate(fn)) == [1, 2, 3]


def test_array_constructor_with_type_mismatch_fails():
    def fn():
        return Array[Simple, 3](1, 2, 3)

    with pytest.raises(TypeError):
        run_and_validate(fn)


def test_array_constructor_with_size_mismatch_fails():
    def fn():
        return Array[int, 3](1, 2)

    with pytest.raises(ValueError, match="should be used with 3 values, got 2"):
        run_and_validate(fn)


def test_array_constructor_with_no_args_fails():
    def fn():
        return Array()

    with pytest.raises(ValueError, match="constructor should be used with at least one value"):
        run_and_validate(fn)


def test_array_constructor_with_heterogeneous_args_fails():
    def fn():
        return Array(1, Simple(2), 3)

    with pytest.raises(TypeError):
        run_and_validate(fn)


def test_array_set():
    def fn():
        array = Array[int, 3](1, 2, 3)
        array[1] = 4
        return array

    assert list(run_and_validate(fn)) == [1, 4, 3]


def test_array_equality():
    def fn():
        a1 = Array(1, 2, 3)
        a2 = Array(1, 2, 3)
        a3 = Array(4, 5, 6)
        a4 = Array(Simple(1), Simple(2), Simple(3))
        a5 = Array(Simple(1), Simple(2), Simple(3))
        a6 = Array(Simple(4), Simple(5), Simple(6))

        assert_true(a1 == a2)
        assert_false(a1 != a2)
        assert_true(a1 != a3)
        assert_false(a1 == a3)
        assert_true(a4 == a5)
        assert_false(a4 != a5)
        assert_true(a4 != a6)
        assert_false(a4 == a6)

        return 1

    assert run_and_validate(fn) == 1


def test_array_equality_of_different_lengths():
    def fn():
        a1 = Array(1, 2, 3)
        a2 = Array(1, 2)

        return a1 == a2

    assert not run_and_validate(fn)


def test_array_record_item_operations():
    def fn():
        array = Array[Simple, 3](Simple(1), Simple(2), Simple(3))
        other1 = Simple(4)
        other2 = Simple(5)

        # Both of these should work the same
        array[1] = other1
        array[2] @= other2

        assert_true(array == Array[Simple, 3](Simple(1), Simple(4), Simple(5)))
        array[1].value = 6
        array[2].value = 7

        assert_true(other1.value == 4)
        assert_true(other2.value == 5)
        assert_true(array[1].value == 6)
        assert_true(array[2].value == 7)

        return array

    assert list(run_and_validate(fn)) == [Simple(1), Simple(6), Simple(7)]


def test_array_contains():
    def fn():
        array = Array(1, 2, 3)

        assert_true(1 in array)
        assert_true(2 in array)
        assert_true(3 in array)
        assert_false(4 in array)

        return 1

    assert run_and_validate(fn) == 1


def test_array_reversed():
    def fn():
        array = Array(1, 2, 3)

        return reversed(array)

    assert list(run_and_validate(fn)) == [3, 2, 1]


def test_array_iteration():
    def fn():
        array = Array(1, 2, 3)
        total = 0

        for i in array:
            total += i

        return total

    assert run_and_validate(fn) == 6


def test_array_enumerate():
    def fn():
        array = Array(1, 3, 5)

        for i, v in enumerate(array):
            assert_true(v == array[i])  # noqa: PLR1736

        return 1

    assert run_and_validate(fn) == 1


def test_array_negative_indexing():
    def fn():
        array = Array(10, 20, 30, 40, 50)

        return Array(array[-1], array[-2], array[-3], array[-4], array[-5])

    assert list(run_and_validate(fn)) == [50, 40, 30, 20, 10]


def test_array_negative_indexing_set():
    def fn():
        array = Array(10, 20, 30, 40, 50)

        array[-1] = 99
        array[-2] = 88
        array[-3] = 77

        return array

    assert list(run_and_validate(fn)) == [10, 20, 77, 88, 99]


@given(args=st.lists(st.integers(min_value=-100, max_value=100), min_size=1, max_size=20))
def test_array_negative_positive_indexing_equivalence(args):
    tuple_args = tuple(args)
    n = len(args)

    def fn():
        array = Array[int, n](*tuple_args)

        results = VarArray[bool, n].new()
        for i in range(n):
            results.append(array[i] == array[i - n])

        return results

    assert all(run_and_validate(fn))


def test_array_index():
    def fn():
        array = Array(1, 2, 3)

        assert_true(array.index(1) == 0)
        assert_true(array.index(2) == 1)
        assert_true(array.index(3) == 2)
        assert_true(array.index(4) == -1)

        return 1

    assert run_and_validate(fn) == 1


def test_array_max():
    def fn():
        array = Array(1, 2, 3)

        return max(array)

    assert run_and_validate(fn) == 3


def test_array_min():
    def fn():
        array = Array(1, 2, 3)

        return min(array)

    assert run_and_validate(fn) == 1


def test_array_count():
    def fn():
        array = Array(1, 2, 3, 2, 1)

        return array.count(1)

    assert run_and_validate(fn) == 2


@given(
    args=st.lists(st.integers(min_value=-9999, max_value=9999), min_size=0, max_size=100),
    reverse=st.booleans(),
)
def test_array_sort(args, reverse: bool):
    tuple_args = tuple(args)
    n = len(args)

    def fn():
        array = Array[int, n](*tuple_args)

        array.sort(reverse=reverse)
        return array

    assert list(run_and_validate(fn)) == sorted(args, reverse=reverse)


@given(
    args=st.lists(st.integers(min_value=-999, max_value=999), min_size=0, max_size=100),
    reverse=st.booleans(),
    a=st.integers(min_value=-9, max_value=9),
    b=st.integers(min_value=-99, max_value=99),
    c=st.integers(min_value=-999, max_value=999),
)
def test_array_sort_with_key(args, reverse: bool, a: int, b: int, c: int):
    tuple_args = tuple(args)
    n = len(args)

    def fn():
        array = Array[int, n](*tuple_args)

        array.sort(key=lambda x: a * x * x + b * x + c, reverse=reverse)
        return array

    assert list(run_and_validate(fn)) == sorted(args, key=lambda x: a * x * x + b * x + c, reverse=reverse)


@given(
    args=st.lists(st.integers(min_value=-999, max_value=999), min_size=1, max_size=100),
    a=st.integers(min_value=-9, max_value=9),
    b=st.integers(min_value=-99, max_value=99),
    c=st.integers(min_value=-999, max_value=999),
)
def test_array_max_with_key(args, a: int, b: int, c: int):
    tuple_args = tuple(args)
    n = len(args)

    def fn():
        array = Array[int, n](*tuple_args)

        return max(array, key=lambda x: a * x * x + b * x + c)

    assert run_and_validate(fn) == max(args, key=lambda x: a * x * x + b * x + c)


@given(
    args=st.lists(st.integers(min_value=-999, max_value=999), min_size=1, max_size=100),
    a=st.integers(min_value=-9, max_value=9),
    b=st.integers(min_value=-99, max_value=99),
    c=st.integers(min_value=-999, max_value=999),
)
def test_array_min_with_key(args, a: int, b: int, c: int):
    tuple_args = tuple(args)
    n = len(args)

    def fn():
        array = Array[int, n](*tuple_args)

        return min(array, key=lambda x: a * x * x + b * x + c)

    assert run_and_validate(fn) == min(args, key=lambda x: a * x * x + b * x + c)


@given(
    args=st.lists(st.integers(min_value=-9999, max_value=9999), min_size=0, max_size=100),
)
def test_array_reverse(args):
    tuple_args = tuple(args)
    n = len(args)

    def fn():
        array = Array[int, n](*tuple_args)

        array.reverse()
        return array

    assert list(run_and_validate(fn)) == list(reversed(args))


class Ele(Record):
    value: int

    def __eq__(self, other):
        return self.value == other.value

    def __ne__(self, other):
        return self.value != other.value

    def __hash__(self):
        return hash(self.value)

    def __lt__(self, other):
        return self.value < other.value

    def __le__(self, other):
        return self.value <= other.value

    def __gt__(self, other):
        return self.value > other.value

    def __ge__(self, other):
        return self.value >= other.value


@given(
    args=st.lists(st.integers(min_value=-9999, max_value=9999), min_size=0, max_size=100),
    reverse=st.booleans(),
)
def test_array_sort_records(args, reverse: bool):
    tuple_args = tuple(Ele(value=v) for v in args)
    n = len(args)

    def fn():
        array = Array[Ele, n](*tuple_args)

        array.sort(reverse=reverse)
        return array

    assert list(run_and_validate(fn)) == sorted(tuple_args, reverse=reverse)


def test_array_truthiness_empty():
    def fn():
        x = Array[int, 0]()
        return 1 if x else 0

    assert run_and_validate(fn) == 0


def test_array_truthiness_non_empty():
    def fn():
        x = Array(1, 2, 3)
        return 1 if x else 0

    assert run_and_validate(fn) == 1
